from typing import List
from unittest.mock import patch
from dataclasses import dataclass, field

import pytest

from intents import Intent, Sys, Entity, follow
from intents.model.intent import IntentParameterMetadata

def test_param_schema_no_params():

    class no_param_intent(Intent):
        """Simple Intent with no parameters"""

    assert no_param_intent.parameter_schema == {}

def test_param_scheme_with_params():

    @dataclass
    class intent_with_params(Intent):
        """Intent with parameters"""
        required_param: Sys.Person
        required_list_param: List[Sys.Person]
        optional_param: Sys.Person = "John"
        optional_list_param: List[Sys.Person] = field(default_factory=lambda: ["Al", "John"])

    assert intent_with_params.parameter_schema == {
        "required_param": IntentParameterMetadata(
            name="required_param",
            entity_cls=Sys.Person,
            is_list=False,
            required=True,
            default=None
        ),
        "required_list_param": IntentParameterMetadata(
            name="required_list_param",
            entity_cls=Sys.Person,
            is_list=True,
            required=True,
            default=None
        ),
        "optional_param": IntentParameterMetadata(
            name="optional_param",
            entity_cls=Sys.Person,
            is_list=False,
            required=False,
            default="John"
        ),
        "optional_list_param": IntentParameterMetadata(
            name="optional_list_param",
            entity_cls=Sys.Person,
            is_list=True,
            required=False,
            default=["Al", "John"]
        ),
    }

def test_param_scheme_invalid_list_default():
    
    with pytest.raises(ValueError):

        class intent_with_invalid_list_default(Intent):
            """Intent with parameters"""
            optional_list_param: List[Sys.Person] = 42

        intent_with_invalid_list_default.parameter_schema

def test_parameter_schema_class_property():
    @dataclass
    class intent_with_params(Intent):
        """Intent with parameters"""
        required_param: Sys.Person
        required_list_param: List[Sys.Person]
        optional_param: Sys.Person = "John"
        optional_list_param: List[Sys.Person] = field(default_factory=lambda: ["Al", "John"])
        
    intent_instance = intent_with_params(required_param=None, required_list_param=None)
    assert intent_with_params.parameter_schema == intent_instance.parameter_schema

def test_parameter_schema_skips_related_intents():
    class PaymentMethod(Entity):
        pass

    @dataclass
    class ask_pay(Intent):
        pass
        
    @dataclass
    class confirm_payment(Intent):
        customer_name: Sys.Person
        contact_numbers: List[Sys.PhoneNumber]
        payment_method: PaymentMethod

        parent_ask_pay: ask_pay = follow()

    expected = {
        'customer_name': IntentParameterMetadata(name='customer_name', entity_cls=Sys.Person, is_list=False, required=True, default=None),
        'contact_numbers': IntentParameterMetadata(name='contact_numbers', entity_cls=Sys.PhoneNumber, is_list=True, required=True, default=None),
        'payment_method': IntentParameterMetadata(name='payment_method', entity_cls=PaymentMethod, is_list=False, required=True, default=None)
    }
    assert confirm_payment.parameter_schema == expected

def test_dataclass_not_applied_twice():
    """https://github.com/dariowho/intents/issues/10"""
    def custom_dataclass(*args, **kwargs):
        custom_dataclass.dataclass_called += 1
    custom_dataclass.dataclass_called = 0

    with patch("intents.model.intent.dataclass", custom_dataclass):
        @custom_dataclass
        class a_class(Intent):
            foo: Sys.Integer = 42
            bar: List[Sys.Integer] = field(default_factory=lambda: [])

        assert custom_dataclass.dataclass_called == 1

def test_warning_if_superclass_not_dataclass():
    @dataclass
    class an_intent(Intent):
        foo: List[Sys.Integer] = field(default_factory=lambda: [])

    class a_sub_intent(an_intent):
        bar: Sys.Integer = 42

    with pytest.warns(None):
        @dataclass
        class a_sub_sub_intent(a_sub_intent):
            babar: Sys.Integer = 43

def test_parent_intents():
    @dataclass
    class BaseIntent(Intent):
        pass

    @dataclass
    class SubIntent(BaseIntent):
        pass

    @dataclass
    class SubSubIntent(SubIntent):
        pass

    assert BaseIntent.parent_intents() == []
    assert SubIntent.parent_intents() == [BaseIntent]
    assert SubSubIntent.parent_intents() == [SubIntent, BaseIntent]

def test_parent_intents__multiple_inheritance():
    @dataclass
    class BaseIntent(Intent):
        pass

    @dataclass
    class OtherBaseIntent(Intent):
        pass

    @dataclass
    class SubIntent(BaseIntent, OtherBaseIntent, int):
        pass

    @dataclass
    class SubSubIntent(SubIntent):
        pass

    assert SubIntent.parent_intents() == [BaseIntent, OtherBaseIntent]
    assert SubSubIntent.parent_intents() == [SubIntent, BaseIntent, OtherBaseIntent]

# def subclass_checks_base_class_parameters():


from unittest.mock import MagicMock
from dataclasses import dataclass
from typing import List

import pytest

from intents import Agent, Intent, Sys
from intents.helpers import coffee_agent as ca
from intents.language import LanguageCode, IntentResponseGroup
from intents.connectors._experimental.snips import prediction_format as pf
from intents.connectors._experimental.snips import prediction, entities

def _get_prediction_component():
    return prediction.SnipsPredictionComponent(ca.CoffeeAgent, entities.ENTITY_MAPPINGS)

def test_prediction_from_parse_renders_language():
    prediction_component = _get_prediction_component()
    prediction_component.intent_from_parse_result = MagicMock(return_value=ca.AskEspresso(roast="dark"))
    parse_result = pf.from_dict({
        'input': 'fake message',
        'intent': {'intentName': 'fake intent', 'probability': 0.677},
        'slots': []
    })
    result = prediction_component.prediction_from_parse_result(parse_result, LanguageCode.ENGLISH)
    assert result.fulfillment_text in ["dark roast espresso, good choice!", "Alright, dark roasted espresso for you"]
    
    messages = result.fulfillment_messages.for_group(IntentResponseGroup.DEFAULT)
    assert messages
    assert messages[0].choices == ["dark roast espresso, good choice!", "Alright, dark roasted espresso for you"]
    
    with pytest.warns(DeprecationWarning):
        messages = result.fulfillment_messages(IntentResponseGroup.DEFAULT)
        assert messages
        assert messages[0].choices == ["dark roast espresso, good choice!", "Alright, dark roasted espresso for you"]

def test_intent_from_parse__default_parameter():
    parse_result = {
        'input': 'I want an espresso',
        'intent': {'intentName': 'AskEspresso', 'probability': 0.677},
        'slots': []
    }
    prediction_component = _get_prediction_component()
    result = prediction_component.intent_from_parse_result(pf.from_dict(parse_result))
    assert result == ca.AskEspresso()


def test_intent_from_parse__single_parameter_single_slot():
    parse_result = {
        'input': 'I want a dark roast espresso',
        'intent': {'intentName': 'AskEspresso', 'probability': 0.56},
        'slots': [
            {
                'range': {'start': 9, 'end': 13},
                'rawValue': 'dark',
                'value': {'kind': 'Custom', 'value': 'dark'},
                'entity': 'CoffeeRoast',
                'slotName': 'roast'
            }
        ]
    }
    prediction_component = _get_prediction_component()
    result = prediction_component.intent_from_parse_result(pf.from_dict(parse_result))
    assert result == ca.AskEspresso(roast="dark")

def test_intent_from_parse__single_parameter_list_slot():
    parse_result = {
        'input': 'I want a dark roast espresso',
        'intent': {'intentName': 'AskEspresso', 'probability': 0.56},
        'slots': [
            {
                'range': {'start': 9, 'end': 13},
                'rawValue': 'dark',
                'value': {'kind': 'Custom', 'value': 'dark'},
                'entity': 'CoffeeRoast',
                'slotName': 'roast'
            },
            {
                'range': {'start': 14, 'end': 19},
                'rawValue': 'light',
                'value': {'kind': 'Custom', 'value': 'light'},
                'entity': 'CoffeeRoast',
                'slotName': 'roast'
            }
        ]
    }
    prediction_component = _get_prediction_component()

    with pytest.warns(None):
        result = prediction_component.intent_from_parse_result(pf.from_dict(parse_result))
    assert result == ca.AskEspresso(roast="dark")

def test_intent_from_parse__list_parameter_list_slot():
    parse_result = {
        'input': 'My colors are green and red',
        'intent': {
            'intentName': 'UserSaysManyColors',
            'probability': 0.66
        },
        'slots': [
            {
                'range': {'start': 14, 'end': 19},
                'rawValue': 'green',
                'value': {'kind': 'Custom', 'value': 'Green'},
                'entity': 'I_IntentsColor',
                'slotName': 'user_color_list'
            }, {
                'range': {'start': 24, 'end': 27},
                'rawValue': 'red',
                'value': {'kind': 'Custom', 'value': 'Red'},
                'entity': 'I_IntentsColor',
                'slotName': 'user_color_list'
            }
        ]
    }

    @dataclass
    class UserSaysManyColors(Intent):
        name = "UserSaysManyColors"
        user_color_list: List[Sys.Color]
    UserSaysManyColors.__intent_language_data__ = ca.mock_language_data(
        UserSaysManyColors,
        ["My colors are $user_color_list{red}, $user_color_list{green} and $user_color_list{blue}", "I give you some colors $user_color_list{yellow}, $user_color_list{orange} and also $user_color_list{purple}"],
        ["$user_color it is"],
        LanguageCode.ENGLISH
    )

    class MyAgent(Agent):
        languages = ['en']
    MyAgent.register(UserSaysManyColors)

    prediction_component = prediction.SnipsPredictionComponent(MyAgent, entities.ENTITY_MAPPINGS)
    
    parse_result = pf.from_dict(parse_result)
    print("NAME", parse_result.intent.intentName)
    result = prediction_component.intent_from_parse_result(parse_result)
    assert result == UserSaysManyColors(user_color_list=['Green', 'Red'])

def test_intent_from_parse__list_parameter_single_slot():
    parse_result = {
        'input': 'My colors are green and red',
        'intent': {
            'intentName': 'UserSaysManyColors',
            'probability': 0.66
        },
        'slots': [
            {
                'range': {'start': 14, 'end': 19},
                'rawValue': 'green',
                'value': {'kind': 'Custom', 'value': 'Green'},
                'entity': 'I_IntentsColor',
                'slotName': 'user_color_list'
            }
        ]
    }

    @dataclass
    class UserSaysManyColors(Intent):
        name = "UserSaysManyColors"
        user_color_list: List[Sys.Color]
    UserSaysManyColors.__intent_language_data__ = ca.mock_language_data(
        UserSaysManyColors,
        ["My colors are $user_color_list{red}, $user_color_list{green} and $user_color_list{blue}", "I give you some colors $user_color_list{yellow}, $user_color_list{orange} and also $user_color_list{purple}"],
        ["$user_color it is"],
        LanguageCode.ENGLISH
    )

    class MyAgent(Agent):
        languages = ['en']
    MyAgent.register(UserSaysManyColors)

    prediction_component = prediction.SnipsPredictionComponent(MyAgent, entities.ENTITY_MAPPINGS)
    
    parse_result = pf.from_dict(parse_result)
    print("NAME", parse_result.intent.intentName)
    result = prediction_component.intent_from_parse_result(parse_result)
    assert result == UserSaysManyColors(user_color_list=['Green'])

#
# Fulfillment
#

def test_fulfillment_recursion_is_blocked():
    # TODO: set timeout on this test

    @dataclass
    class RecursiveFulfillmentIntent(Intent):
        def fulfill(self, context):
            return RecursiveFulfillmentIntent()

    RecursiveFulfillmentIntent.__intent_language_data__ = ca.mock_language_data(
        RecursiveFulfillmentIntent, [], ["fake response"], LanguageCode.ENGLISH
    )

    class MyAgent(Agent):
        languages = ['en']
    MyAgent.register(RecursiveFulfillmentIntent)

    prediction_component = prediction.SnipsPredictionComponent(MyAgent, entities.ENTITY_MAPPINGS)
    pred = prediction_component.prediction_from_intent(RecursiveFulfillmentIntent(), LanguageCode.ENGLISH)
    with pytest.raises(RecursionError):
        prediction_component.fulfill_local(pred, LanguageCode.ENGLISH)



"""
Alexa has peculiar requirements about language. Here we convert language data in
a format that Alexa can digest.
"""
import re
from dataclasses import dataclass
from collections import defaultdict
from typing import Dict, List, Type

from intents import EntityMixin, Agent, LanguageCode
from intents.language import entity_language, match_agent_language

LOCALE_MAP = {
    "ar-SA": None,
    "de-DE": LanguageCode.GERMAN,
    "en-AU": LanguageCode.ENGLISH,
    "en-CA": LanguageCode.ENGLISH_US,
    "en-GB": LanguageCode.ENGLISH_UK,
    "en-IN": LanguageCode.ENGLISH,
    "en-US": LanguageCode.ENGLISH_US,
    "es-ES": LanguageCode.SPANISH,
    "es-MX": LanguageCode.SPANISH_LATIN_AMERICA,
    "es-US": LanguageCode.SPANISH_LATIN_AMERICA,
    "fr-CA": LanguageCode.FRENCH,
    "fr-FR": LanguageCode.FRENCH,
    "hi-IN": None,
    "it-IT": LanguageCode.ITALIAN,
    "ja-JP": None,
    "pt-BR": None
}

@dataclass
class AlexaEntityEntry(entity_language.EntityEntry):
    """
    Not all entity entry values/synonyms can be exported to Alexa: those which
    are not acceptable are must be stripped from result. Here we save
    Alexa-acceptable values in two additional fields.
    """
    alexa_value: str = None
    alexa_synonyms: List[str] = None

class AlexaLanguageComponent:

    agent_cls: Type[Agent]

    _entry_id_to_value: Dict[LanguageCode, Dict[str, str]]

    def __init__(self, agent_cls: Type[Agent]):
        self.agent_cls = agent_cls
        self._entry_id_to_value = {}

        self._build_indices(agent_cls)

    def _build_indices(self, agent_cls: Type[Agent]):
        # TODO: optimize (this loads all language data for entities)
        self._entry_id_to_value = defaultdict(dict)
        for entity_cls in agent_cls._entities_by_name.values():
            language_data = self._entity_language_data(agent_cls, entity_cls)
            for language, entries in language_data.items():
                entries_to_value = {}
                for entry in entries:
                    value_id = self.entry_value_id(entity_cls, entry)
                    entries_to_value[value_id] = entry.value
                self._entry_id_to_value[language].update(entries_to_value)
        self._entry_id_to_value = dict(self._entry_id_to_value)

    def alexa_locale_to_agent_language(self, locale_str: str) -> LanguageCode:
        """
        Converts a Locale, as it comes in Alexa requests, to one of the
        languages that are supported by Agent. The following scenarios may
        happen:

        * Alexa locale is not supported by *Intents* (e.g. "ar-SA") ->
          :class:`KeyError`
        * Alexa locale is one of the Agent supported languages ->
          :class:`LanguageCode` returned
        * Alexa locale is not in Agent supported languages, but there's a
          fallback available (e.g. "en-US" can fallback on "en" or "en-GB") ->
          :class:`LanguageCode` returned
        """
        language_code = LOCALE_MAP[locale_str]
        if not language_code:
            raise KeyError(f"Locale {locale_str} is not supported by Intents")

        return match_agent_language(self.agent_cls, language_code)

    def alexa_entry_id_to_value(self, alexa_entry_id: str, lang: LanguageCode) -> str:
        """
        Alexa returns an entry ID for each custom entry match. We define entry
        IDs in export, and here we translate them back to the original entry
        value, which is typically needed in prediction.

        Note that the entry value that Alexa knows may be different from the
        "canonical" value that is defined in language files. This is because
        some values/synonyms cannot be exported to Alexa. For instance, let's
        consider the :class:`CalculatorOperator` entity, with the following
        language data:

        .. code-block:: yaml

            entries:
                "*":
                    - times
                    - multiplied by
                    - product
                "+":
                    - plus
                    - addition
                ...

        Here `*` is not a valid value for Alexa. So we'll use the first viable
        synonim (`"times"`, which will produce an entry id like
        `"CalculatorOperator-times"`) as main value for Alexa, and the remaining
        ones as synonyms (`["multiplied by", "product"]`). However, we want to
        get back the original `*` to caller during inference, and this is what
        this method does:

        >>> lc.alexa_entry_id_to_value("CalculatorOperator-times", LanguageCode.ENGLISH)
        "*"
        """
        return self._entry_id_to_value[lang][alexa_entry_id]

    def entity_language_data(self, entity_cls: Type[EntityMixin], lang: LanguageCode=None):
        return self._entity_language_data(self.agent_cls, entity_cls, lang)

    @staticmethod
    def _entity_language_data(
        agent_cls: Type[Agent],
        entity_cls: Type[EntityMixin],
        lang: LanguageCode=None
    ) -> Dict[LanguageCode, List[AlexaEntityEntry]]:
        """
        Load Entity language data, removing values and synonyms that can't be
        exported to Alexa.
        """
        result = {}
        language_data = entity_language.entity_language_data(agent_cls, entity_cls, lang)
        for language_code, entries in language_data.items():
            alexa_entries = []
            for entry in entries:
                all_values = [entry.value] + entry.synonyms
                acceptable_values = [v for v in all_values if _is_valid_entity_value(v)]
                if not acceptable_values:
                    raise ValueError(f"Entry {entry} of custom entity {entity_cls} has no valid values for "
                                     "Alexa. Please include at leas one standard alphanumeric synonym")
                alexa_entries.append(AlexaEntityEntry(
                    value=entry.value,
                    synonyms=entry.synonyms,
                    alexa_value=acceptable_values[0],
                    alexa_synonyms=acceptable_values[1:]
                ))
            result[language_code] = alexa_entries
        return result

    @staticmethod
    def entry_value_id(entity_cls: Type[EntityMixin], entry: AlexaEntityEntry):
        """
        Entity entries in Alexa have IDs. This is a centralized function to
        compute them.
        """
        return entity_cls.name + "-" + entry.alexa_value.replace(" ", "") # TODO: refine


def _is_valid_entity_value(val: str):
    # TODO: refine
    return bool(re.match(r'^[a-zA-Z0-9\s]+$', val))



import pytest

from intents.model.names import check_name

def test_check_name():
    check_name("valid_intent_name")
    check_name("AlsoValidName")
    check_name("valid.NameAsWell")
    with pytest.raises(ValueError):
        check_name("Invalid Name")
    with pytest.raises(ValueError):
        check_name("invalid $name")
    with pytest.raises(ValueError):
        check_name("invalid__Name")
    with pytest.raises(ValueError):
        check_name("_InvalidName")
    with pytest.raises(ValueError):
        check_name("i_InvalidName")



"""
Here we convert Intent object names into names that will be used in Alexa
"""
from typing import Dict, Type

from intents import Intent, Agent, EntityMixin
from intents.connectors._experimental.alexa import slot_types

class AlexaNamesComponent:
    agent_cls: Type[Agent]

    _alexa_to_intent_name: Dict[str, str]
    _alexa_entry_id_to_canonical: Dict[str, str]

    def __init__(self, agent_cls: Type[Agent]):
        self.agent_cls = agent_cls
        self._alexa_to_intent_name = {}

        self._build_indices(agent_cls)

    def _build_indices(self, agent_cls: Type[Agent]):
        for intent_cls in agent_cls.intents:
            alexa_name = self.intent_to_alexa(intent_cls)
            self._alexa_to_intent_name[alexa_name] = intent_cls.name

    def alexa_to_intent_name(self, alexa_name: str) -> str:
        """
        Convert an Intent name, as it is referenced in the Alexa agent
        """
        return self._alexa_to_intent_name[alexa_name]

    @staticmethod
    def intent_to_alexa(intent_cls: Type[Intent]) -> str:
        """
        Generate the Intent name that will be used in the Alexa export
        """
        return intent_cls.name.replace(".", "_")

    @staticmethod
    def entity_service_name(entity_cls: Type[EntityMixin]):
        # TODO: may be necessary to sanitize custom entity names
        return slot_types.ENTITY_MAPPINGS.service_name(entity_cls)



"""
Here we implement :class:`DialogflowEsConnector`, an implementation of
:class:`Connector` that allows Agents to operate on Dialogflow ES.
"""
import os
import logging
import tempfile
from dataclasses import dataclass, field
from typing import Set, Dict, Union, Iterable, Type

import google.auth.credentials
from google.cloud.dialogflow_v2.types import TextInput, QueryInput, EventInput
from google.cloud.dialogflow_v2.services.sessions import SessionsClient
from google.cloud.dialogflow_v2.services.agents import AgentsClient
from google.cloud.dialogflow_v2 import types as pb

from intents import Agent, Intent, LanguageCode, FulfillmentContext, FulfillmentResult
from intents.model.relations import intent_relations
from intents.language_codes import ensure_language_code
from intents.connectors.interface import Connector, Prediction, FulfillmentRequest, WebhookConfiguration, deserialize_intent_parameters
from intents.connectors.dialogflow_es.auth import resolve_credentials
from intents.connectors.dialogflow_es.util import dict_to_protobuf
from intents.connectors.dialogflow_es import webhook
from intents.connectors.dialogflow_es import names as df_names
from intents.connectors.dialogflow_es import entities as df_entities
from intents.connectors.dialogflow_es import export as df_export
from intents.connectors.dialogflow_es.prediction import PredictionBody, DetectIntentBody, WebhookRequestBody, intent_responses

logger = logging.getLogger(__name__)

RICH_RESPONSE_PLATFORMS = ["telegram", "facebook", "slack", "line", "hangouts"]

# Dialogflow makes use of Protobuffer for its data structures, and protobuf may be
# tricky to deal with. For instance, `MessageToDict` will convert snake_case to
# lowerCamelCase, so API is documented snake_case, protobuf is snake_case,
# REST dialogflow results are camelCase, protobuf converted to dict is camelCase
# (unless you use a flag, in that case it could also be snake_case) 💀 This is one
# of the reasons this library exists.

@dataclass
class DialogflowPrediction(Prediction):
    """
    This is an implementation of :class:`~intents.connectors.interface.Prediction`
    that comes from Dialogflow. It adds a `df_response` field, through which the
    full Dialogflow prediction payload can be accessed. Note that this is a tool
    for debugging: relying on Dialogflow data in your business logic is likely
    to make it harder to connect your Agent to different platforms.

    `DialogflowPredictions` are produced internally by
    :class:`DialogflowEsConnector`, and are returned by its :meth:`~DialogflowEsConnector.predict` and
    :meth:`~DialogflowEsConnector.trigger` methods.

    Args:
        intent: An instance of the predicted Intent
        confidence: Dialogflow's confidence on its prediction
        fulfillment_messages: A map of Intent Responses, as they were
            returned by the Service
        fulfillment_text: A plain-text version of the response
        df_response: Raw Dialogflow response data. It is advisable not to rely
            on this is production, if you want to keep cross-service compatibility
    """
    df_response: DetectIntentBody = field(default=False, repr=False)

class DialogflowEsConnector(Connector):
    """
    This is an implementation of :class:`~intents.connectors.interface.Connector`
    that enables Agents to work as Dialogflow projects.

    An Agent can be connected to Dialogflow by providing its :class:`~intents.model.agent.Agent`
    class and service account credentials for the the Google Cloud project
    that hosts the Dialogflow ES agent:

    .. code-block:: python

        from example_agent import ExampleAgent
        from intents.connectors import DialogflowEsConnector
        df = DialogflowEsConnector('/path/to/your/service-account-credentials.json', ExampleAgent)

    The Connector can now be used, mainly to

    * Export the Agent with :meth:`DialogflowEsConnector.export`
    * Predict an utterance with :meth:`DialogflowEsConnector.predict`
    * Trigger an Intent with :meth:`DialogflowEsConnector.trigger`

    Args:
        google_credentials: Path to service account JSON credentials, or a Credentials object
        agent_cls: The Agent to connect
        default_session: An arbitrary string to identify the conversation during
            predictions. If None, Connector will generate a random string
        default_language: Default language to use during predictions. If None, Connector
            will use the Agent's firs defined language.
        rich_platforms: Platforms to include when exporting Rich response messages
        webhook_configuration: Webhook connection parameters
    """
    entity_mappings = df_entities.MAPPINGS
    rich_platforms: Iterable[str]
    webhook_configuration: WebhookConfiguration

    _credentials: google.auth.credentials.Credentials
    _session_client: SessionsClient
    _need_context_set: Set[type(Intent)]
    _intents_by_context: Dict[str, type(Intent)]

    def __init__(
        self,
        google_credentials: Union[str, google.auth.credentials.Credentials],
        agent_cls: type(Agent),
        default_session: str=None,
        default_language: Union[LanguageCode, str]=None,
        rich_platforms: Iterable[str]=("telegram",),
        webhook_configuration: WebhookConfiguration=None
    ):
        super().__init__(agent_cls, default_session=default_session,
                         default_language=default_language)
        self._credentials = resolve_credentials(google_credentials)
        assert all([p in RICH_RESPONSE_PLATFORMS for p in rich_platforms])
        self._session_client = SessionsClient(credentials=self._credentials)
        self.rich_platforms = rich_platforms
        self.webhook_configuration = webhook_configuration
        self._need_context_set = _build_need_context_set(agent_cls)
        self._intents_by_context = _build_intents_by_context(agent_cls)

    @property
    def gcp_project_id(self) -> str:
        """
        Return the Google Cloud Project ID that is associated with the current Connection
        """
        return self._credentials.project_id

    def export(self, destination: str):
        agent_name = 'py-' + self.agent_cls.__name__
        return df_export.export(self, destination, agent_name)

    def upload(self):
        agents_client = AgentsClient(credentials=self._credentials)
        with tempfile.TemporaryDirectory() as tmp_dir:
            export_path = os.path.join(tmp_dir, 'agent.zip')
            self.export(export_path)
            with open(export_path, 'rb') as f:
                agent_content = f.read()
            restore_request = pb.RestoreAgentRequest(
                parent=f"projects/{self.gcp_project_id}",
                agent_content=agent_content
            )
            agents_client.restore_agent(request=restore_request)

    def predict(self, message: str, session: str = None, language: Union[LanguageCode, str] = None) -> DialogflowPrediction:
        if not session:
            session = self.default_session
        if not language:
            language = self.default_language

        language = ensure_language_code(language)
        text_input = TextInput(text=message, language_code=language.value)
        query_input = QueryInput(text=text_input)
        session_path = self._session_client.session_path(
            self.gcp_project_id, session)
        df_result = self._session_client.detect_intent(
            session=session_path,
            query_input=query_input
        )
        df_response = DetectIntentBody(df_result)

        return self._df_body_to_prediction(df_response)

    def trigger(self, intent: Intent, session: str=None, language: Union[LanguageCode, str]=None) -> DialogflowPrediction:
        if not session:
            session = self.default_session
        if not language:
            language = self.default_language

        language = ensure_language_code(language)
        intent_name = intent.name
        event_name = df_names.event_name(intent.__class__)
        event_parameters = {}
        for param_name, param_metadata in intent.parameter_schema.items():
            param_mapping = df_entities.MAPPINGS[param_metadata.entity_cls]
            if param_name in intent.__dict__:
                param_value = intent.__dict__[param_name]
                event_parameters[param_name] = param_mapping.to_service(param_value)

        logger.info("Triggering event '%s' in session '%s' with parameters: %s",
                    event_name, session, event_parameters)
        if not event_parameters:
            event_parameters = {}

        event_input = EventInput(
            name=event_name,
            parameters=dict_to_protobuf(event_parameters),
            language_code=language.value
        )
        query_input = QueryInput(event=event_input)
        session_path = self._session_client.session_path(
            self.gcp_project_id, session)
        df_result = self._session_client.detect_intent(
            session=session_path,
            query_input=query_input
        )
        df_response = DetectIntentBody(df_result)

        return self._df_body_to_prediction(df_response)

    def fulfill(self, fulfillment_request: FulfillmentRequest) -> dict:
        webhook_body = WebhookRequestBody(fulfillment_request.body)
        intent = self._df_body_to_intent(webhook_body)
        context = self._df_body_to_fulfillment_context(webhook_body)
        fulfillment_result = FulfillmentResult.ensure(intent.fulfill(context))
        logger.debug("Returning fulfillment result: %s", fulfillment_result)
        if fulfillment_result:
            return webhook.fulfillment_result_to_response(fulfillment_result, context)
        return {}

    def _df_body_to_fulfillment_context(self, df_body: DetectIntentBody) -> DialogflowPrediction:
        return FulfillmentContext(
            confidence=df_body.queryResult.intentDetectionConfidence,
            fulfillment_messages=intent_responses(df_body),
            fulfillment_text=df_body.queryResult.fulfillmentText,
            language=LanguageCode(df_body.queryResult.languageCode)
        )

    def _df_body_to_prediction(self, df_body: DetectIntentBody) -> DialogflowPrediction:
        return DialogflowPrediction(
            intent=self._df_body_to_intent(df_body),
            confidence=df_body.queryResult.intentDetectionConfidence,
            fulfillment_messages=intent_responses(df_body),
            fulfillment_text=df_body.queryResult.fulfillmentText,
            df_response=df_body.detect_intent
        )

    def _df_body_to_intent(
        self,
        df_body: PredictionBody,
        build_related_cls: Type[Intent]=None,
        visited_intents: Set[Type[Intent]]=None
    ) -> Intent:
        """
        Convert a Dialogflow prediction response into an instance of
        :class:`Intent`.
        
        This method is recursive on intent relations. When an intent has a
        :meth:`~intents.model.relations.follow` field, that field must be filled
        with an instance of the followed intent; in this case
        :meth:`_df_body_to_intent` will call itself passing the parent intent
        class as `build_related_cls`, to force building that intent from the
        same `df_body`; contexts and parameters will be checked for consistency.

        Args:
            df_body: A Dialogflow Response
            build_related_cls: Force to build the related intent instead of
                the predicted one
            visited_intents: This is used internally to prevent recursion loops
        """
        if not visited_intents:
            visited_intents = set()

        contexts, context_parameters = df_body.contexts()

        # Slot filling in progress
        # TODO: also check queryResult.cancelsSlotFilling
        # if "__system_counters__" in contexts:
        if not df_body.queryResult.allRequiredParamsPresent:
            logger.warning("Prediction doesn't have values for all required parameters. "
                           "Slot filling may be in progress, but this is not modeled yet: "
                           "Intent object will be None")
            return None

        if build_related_cls:
            # TODO: adjust lifespan
            intent_cls = build_related_cls
            df_parameters = {
                p_name: p.value for p_name, p in context_parameters.items() 
                if p_name in intent_cls.parameter_schema
            }
        else:
            intent_name = df_body.intent_name
            intent_cls: Intent = self.agent_cls._intents_by_name.get(intent_name)
            if not intent_cls:
                raise ValueError(f"Prediction returned intent '{intent_name}', " +
                    "but this was not found in Agent definition. Make sure to restore a latest " +
                    "Agent export from `services.dialogflow_es.export.export()`. If the problem " +
                    "persists, please file a bug on the Intents repository.")
            df_parameters = df_body.intent_parameters

        visited_intents.add(intent_cls)
        parameter_dict = deserialize_intent_parameters(df_parameters, intent_cls, self.entity_mappings)
        related_intents_dict = {}
        for rel in intent_relations(intent_cls).follow:
            if rel.target_cls in visited_intents:
                raise ValueError(f"Loop detected: {rel.target_cls} was already visited. Make sure "
                                 "your Agent has no circular dependencies")
            related_intent = self._df_body_to_intent(df_body, rel.target_cls, visited_intents)
            related_intent.lifespan = df_body.context_lifespans.get(df_names.context_name(rel.target_cls), 0)
            
            related_intents_dict[rel.field_name] = related_intent

        result = intent_cls(**parameter_dict, **related_intents_dict)
        result.lifespan = df_body.context_lifespans.get(df_names.context_name(intent_cls), 0)
        return result

    def _intent_needs_context(self, intent: Intent) -> bool:
        return intent in self._need_context_set

def _build_need_context_set(agent_cls: type(Agent)) -> Set[Intent]:
    """
    Return a list of intents that need to spawn a context, based on their
    relations:

    * If intent *B* follows intent *A*, then intent *A* needs to spawn a context
    """
    result = set()
    for intent in agent_cls.intents:
        related = intent_relations(intent)
        for rel in related.follow:
            result.add(rel.target_cls)
    return result

def _build_intents_by_context(agent_cls: Type[Agent]) -> Dict[str, Type[Intent]]:
    result = {}
    for intent_cls in agent_cls.intents:
        context_name = df_names.context_name(intent_cls)
        if context_name in result:
            raise ValueError(f"Intents '{intent_cls.name}' and '{result[context_name].name}' " +
                "have ambiguous context name. This is a bug: please file an issue on the " +
                "Intents repo. Quick fix: change the name of one of the two intents.")
        result[context_name] = intent_cls
    return result



"""
The `AMAZON.DATE` slot type in Alexa may return non-standard date formats. Here
we parse them to Python :class:`date` objects. More details at
https://developer.amazon.com/en-US/docs/alexa/custom-skills/slot-type-reference.html#date

.. info::

    Differently from other Services, dates in Alexa can either be specific dates
    (e.g. "2021-07-11") or date intervals (e.g. "2021-W27-WE"). To keep behavior
    consistent we return the first day of the interval. An improvement for the
    future could be to define specific Alexa system entities, to be used in Agents
    that don't need to be compatible with other services.

"""
import re
import calendar
from typing import Tuple, List
from datetime import date, timedelta

SEASONS = {
    "SP": {
        "from": (3, 21),
        "to": (6, 21)
    },
    "SU": {
        "from": (6, 22),
        "to": (9, 22)
    },
    "FA": {
        "from": (9, 23),
        "to": (12, 21)
    },
    "WI": {
        "from": (12, 22),
        "to": (3, 20)
    }
}

def parse_alexa_date(alexa_date: str) -> Tuple[date, date]:
    """
    Parse an Alexa date string, as it is serialized by the `AMAZON.DATE` slot
    type.
    
    Return `date_from` and `date_to` :class:`date` objects. If `alexa_date` is a
    specific date, then `date_to` will be `None`.

    First day of the week is Monday.
    """
    date_from = None
    date_to = None

    # 2015-11-25
    match = re.match(r"^(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})$", alexa_date)
    if match:
        year, month, day = _integer_groups(match, ["year", "month", "day"])
        date_from = date(year, month, day)
        
    # 2015-W49
    match = re.match(r"^(?P<year>\d{4})-W(?P<week>\d{1,2})$", alexa_date)
    if match:
        year, week = _integer_groups(match, ["year", "week"])
        date_from = date.fromisocalendar(year, week, 1)
        date_to = date_from + timedelta(days=6)

    # 2015-W49-WE
    match = re.match(r"^(?P<year>\d{4})-W(?P<week>\d{1,2})-WE$", alexa_date)
    if match:
        year, week = _integer_groups(match, ["year", "week"])
        monday = date.fromisocalendar(year, week, 1)
        date_from = monday + timedelta(days=5)
        date_to = monday + timedelta(days=6)

    # 2018-09 / 2018-09-XX
    match = re.match(r"^(?P<year>\d{4})-(?P<month>\d{1,2})(?:-XX)?$", alexa_date)
    if match:
        year, month = _integer_groups(match, ["year", "month"])
        _, last_day = calendar.monthrange(year, month)
        date_from = date(year, month, 1)
        date_to = date_from.replace(day=last_day)

    # 2018 / 2018-XX-XX
    match = re.match(r"^(?P<year>\d{4})(?:-XX-XX)?$", alexa_date)
    if match:
        year, = _integer_groups(match, ["year"])
        date_from = date(year, 1, 1)
        date_to = date(year, 12, 31)

    # 197X
    match = re.match(r"^(?P<decade>\d{3})X$", alexa_date)
    if match:
        decade, = _integer_groups(match, ["decade"])
        year = decade * 10
        date_from = date(year, 1, 1)
        date_to = date(year+9, 12, 31)

    # 2021-SP / 2021-SU / 2021-FA / 2021-WI
    match = re.match(r"(?P<year>\d{4})-(?P<season>SP|SU|FA|WI)$", alexa_date)
    if match:
        from_year, = _integer_groups(match, ["year"])
        season = match.group("season")
        from_month, from_day = SEASONS[season]["from"]
        to_month, to_day = SEASONS[season]["to"]
        to_year = from_year + 1 if season == "WI" else from_year
        date_from = date(from_year, from_month, from_day)
        date_to = date(to_year, to_month, to_day)

    if date_from:
        return date_from, date_to
    else:
        raise ValueError(f"Could not parse Alexa date string: {alexa_date}. This is possibly a bug, please file an issue at https://github.com/dariowho/intents")

def _integer_groups(match: re.Match, groups=List[str]) -> List[int]:
    result = []
    for g in groups:
        result.append(int(match.group(g)))
    return result



import tempfile

from example_agent.agent import ExampleAgent
from intents.connectors._experimental.alexa import AlexaConnector

def test_export_example_agent_no_exceptions():
    alexa = AlexaConnector(ExampleAgent, "any invocation")
    with tempfile.TemporaryDirectory() as temp_dir:
        alexa.export(temp_dir)



from dataclasses import dataclass
from unittest.mock import patch

import pytest
from google.cloud.dialogflow_v2.types import DetectIntentResponse

from intents.connectors.dialogflow_es.connector import DialogflowEsConnector
from intents import language, Agent, Intent, Entity, LanguageCode, Sys
from intents.connectors.interface import Prediction, FulfillmentRequest
from intents.helpers import coffee_agent
from example_agent import ExampleAgent, travels

# A full Dialogflow Response from ExampleAgent, with
# - A text message in the DEFAULT platform
# - A text message in the TELEGRAM platform
# - A quick replies message in the TELEGRAM platform
df_response_quick_replies_serialized = b'\n-1cedb9e6-f958-437f-9299-74f966fbec62-9779ea79\x12\xa2\x03\n\x10i want to travel"\x00(\x012QIf you like I can recommend you an hotel. Or I can send you some holiday pictures:U\nS\nQIf you like I can recommend you an hotel. Or I can send you some holiday pictures:.\n*\n(I also like travels, how can I help you?0\x03:;\x1a7\n\rQuick Replies\x12\x12Recommend an hotel\x12\x12Send holiday photo0\x03Zj\nOprojects/learning-dialogflow/agent/intents/e3a1e749-be67-11eb-8ad8-bbef97dc13e7\x12\x17travels.UserWantsTravele\x00\x00\x80?z\x02en'
df_response_quick_replies = DetectIntentResponse.deserialize(df_response_quick_replies_serialized)

# CoffeeAgent "I'd like an espresso"
df_response_espresso_serialized = b'\n-8939a534-88f9-4e49-8ad1-eaec18543743-c4f60134\x12\xc4\x03\n\x14I\'d like an espresso"\x13\n\x11\n\x05roast\x12\x08\x1a\x06medium(\x012\x0cAny response:\x10\n\x0e\n\x0cAny responseR\x80\x01\nQprojects/learning-dialogflow/agent/sessions/testing-session/contexts/c_ask_coffee\x10\x05\x1a)\n\x11\n\x05roast\x12\x08\x1a\x06medium\n\x14\n\x0eroast.original\x12\x02\x1a\x00R\x88\x01\nYprojects/learning-dialogflow/agent/sessions/testing-session/contexts/c_testing_ask_coffee\x10\x04\x1a)\n\x11\n\x05roast\x12\x08\x1a\x06medium\n\x14\n\x0eroast.original\x12\x02\x1a\x00Z^\nOprojects/learning-dialogflow/agent/intents/9d3fe183-efd1-11eb-a79e-17bc86f5a601\x12\x0bAskEspressoe\x00\x00\x80?z\x02en'
df_response_espresso = DetectIntentResponse.deserialize(df_response_espresso_serialized)

# CoffeeAgent "I'd like an espresso" > "With milk"
df_response_espresso_milk_serialized = b'\n-2478f1c9-6549-4f7d-90c8-9fd493f66a42-c4f60134\x12\xf7\x03\n\tWith milk"\x00(\x012\x0cAny response:\x10\n\x0e\n\x0cAny responseRS\nOprojects/learning-dialogflow/agent/sessions/testing-session/contexts/c_add_milk\x10\x05R\x80\x01\nQprojects/learning-dialogflow/agent/sessions/testing-session/contexts/c_ask_coffee\x10\x04\x1a)\n\x14\n\x0eroast.original\x12\x02\x1a\x00\n\x11\n\x05roast\x12\x08\x1a\x06mediumR\x88\x01\nYprojects/learning-dialogflow/agent/sessions/testing-session/contexts/c_testing_ask_coffee\x10\x03\x1a)\n\x11\n\x05roast\x12\x08\x1a\x06medium\n\x14\n\x0eroast.original\x12\x02\x1a\x00ZZ\nOprojects/learning-dialogflow/agent/intents/9d3fe186-efd1-11eb-a79e-17bc86f5a601\x12\x07AddMilke\x00\x00\x80?z\x02en'
df_response_espresso_milk = DetectIntentResponse.deserialize(df_response_espresso_milk_serialized)

# CoffeeAgent "I'd like an espresso" > "With milk" > "And no foam"
df_response_espresso_milk_nofoam_serialized = b'\n-05c5d723-2f30-4d65-a590-6429a445144b-c4f60134\x12\xfb\x03\n\x0bAnd no foam"\x00(\x012\x0cAny response:\x10\n\x0e\n\x0cAny responseR\x80\x01\nQprojects/learning-dialogflow/agent/sessions/testing-session/contexts/c_ask_coffee\x10\x03\x1a)\n\x11\n\x05roast\x12\x08\x1a\x06medium\n\x14\n\x0eroast.original\x12\x02\x1a\x00R\x88\x01\nYprojects/learning-dialogflow/agent/sessions/testing-session/contexts/c_testing_ask_coffee\x10\x02\x1a)\n\x14\n\x0eroast.original\x12\x02\x1a\x00\n\x11\n\x05roast\x12\x08\x1a\x06mediumRS\nOprojects/learning-dialogflow/agent/sessions/testing-session/contexts/c_add_milk\x10\x04Z\\\nOprojects/learning-dialogflow/agent/intents/9d3fe18a-efd1-11eb-a79e-17bc86f5a601\x12\tAndNoFoame\x00\x00\x80?z\x02en'
df_response_espresso_milk_nofoam = DetectIntentResponse.deserialize(df_response_espresso_milk_nofoam_serialized)

#
# Tests
#

@patch("intents.connectors.dialogflow_es.connector.resolve_credentials")
@patch("intents.connectors.dialogflow_es.connector.SessionsClient")
def test_predict(mock_df_client_class, *args):
    # TODO: this relies on the consistency between mock prediction and
    # ExampleAgent, consider switching to CoffeeAgent
    # TODO: this is more functional test than unit test...
    def mock_df_client():
        pass
    def mock_detect_intent(session, query_input):
        return df_response_quick_replies
    def mock_session_path(gcp_project_id, session):
        return 'fake-full-session-path'
    mock_df_client.detect_intent = mock_detect_intent
    mock_df_client.session_path = mock_session_path
    mock_df_client_class.return_value = mock_df_client

    df = DialogflowEsConnector('/fake/path/to/credentials.json', ExampleAgent)
    predicted = df.predict("A fake sentence")
    expected_responses = {
        language.IntentResponseGroup.DEFAULT: [
            language.TextIntentResponse(choices=["If you like I can recommend you an hotel. Or I can send you some holiday pictures"])
        ],
        language.IntentResponseGroup.RICH: [
            language.TextIntentResponse(choices=["I also like travels, how can I help you?"]),
            language.QuickRepliesIntentResponse(replies=["Recommend an hotel", "Send holiday photo"])
        ]
    }
    assert isinstance(predicted.intent, travels.UserWantsTravel)
    # pylint: disable=no-member # (protobuf...)
    assert predicted.fulfillment_text == df_response_quick_replies.query_result.fulfillment_text
    assert predicted.fulfillment_messages == expected_responses

    with pytest.warns(DeprecationWarning):
        assert predicted.fulfillment_message_dict == expected_responses

@patch("intents.connectors.dialogflow_es.connector.resolve_credentials")
@patch("intents.connectors.dialogflow_es.connector.SessionsClient")
def test_predict_related_intents_follow(mock_df_client_class, *args):
    def mock_df_client():
        pass
    def mock_detect_intent(session, query_input):
        return df_response_espresso_milk_nofoam # AskEspresso > WithMilk > AndNoFoam
    def mock_session_path(gcp_project_id, session):
        return 'testing-session'
    mock_df_client.detect_intent = mock_detect_intent
    mock_df_client.session_path = mock_session_path
    mock_df_client_class.return_value = mock_df_client

    df = DialogflowEsConnector('/fake/path/to/credentials.json', coffee_agent.CoffeeAgent)
    predicted = df.predict("A fake sentence")
    intent: coffee_agent.AndNoFoam = predicted.intent
    assert isinstance(intent, coffee_agent.AndNoFoam)
    assert isinstance(intent.parent_add_milk, coffee_agent.AddMilk)
    assert isinstance(intent.parent_add_milk.parent_ask_coffee, coffee_agent.AskCoffee) # TODO: or should be AskEspresso?
    assert intent.parent_add_milk.parent_ask_coffee.roast == "medium"
    assert intent.lifespan == 0 # No followups -> does not spawn a context
    assert intent.parent_add_milk.lifespan == 4
    assert intent.parent_add_milk.parent_ask_coffee.lifespan == 3

@patch("intents.connectors.dialogflow_es.connector.resolve_credentials")
def test_intent_need_context(m_credentials):
    df = DialogflowEsConnector('/fake/path/to/credentials.json', coffee_agent.CoffeeAgent)
    assert df._intent_needs_context(coffee_agent.AskCoffee) == True
    assert df._intent_needs_context(coffee_agent.AskEspresso) == False
    assert df._intent_needs_context(coffee_agent.AddMilk) == True
    assert df._intent_needs_context(coffee_agent.AddSkimmedMilk) == False
    assert df._intent_needs_context(coffee_agent.AndNoFoam) == False
#
# Prediction
#

def test_fulfillment_messages():
    mock_default_messages = [
        language.TextIntentResponse(choices=["If you like I can recommend you an hotel. Or I can send you some holiday pictures"])
    ]
    mock_rich_messages = [
        language.TextIntentResponse(choices=["I also like travels, how can I help you?"]),
        language.QuickRepliesIntentResponse(replies=["Recommend an hotel", "Send holiday photo"])
    ]
    prediction = Prediction(
        intent=None,
        confidence=0.5,
        fulfillment_messages=language.IntentResponseDict({
            language.IntentResponseGroup.DEFAULT: mock_default_messages,
            language.IntentResponseGroup.RICH: mock_rich_messages
        }),
        fulfillment_text="Fake fulfillment text"
    )

    assert prediction.fulfillment_messages.for_group() == mock_rich_messages
    assert prediction.fulfillment_messages.for_group(language.IntentResponseGroup.DEFAULT) == mock_default_messages
    assert prediction.fulfillment_messages.for_group(language.IntentResponseGroup.RICH) == mock_rich_messages

    with pytest.warns(DeprecationWarning):
        assert prediction.fulfillment_messages() == mock_rich_messages
        assert prediction.fulfillment_messages(language.IntentResponseGroup.DEFAULT) == mock_default_messages
        assert prediction.fulfillment_messages(language.IntentResponseGroup.RICH) == mock_rich_messages

#
# Fulfillment
#

# "How much is three times five?"
CALCULATOR_FULFILLMENT = {"responseId": "fake-response-id", "queryResult": {"queryText": "how much is two plus 2?", "parameters": {"first_operand": 2.0, "second_operand": 2.0, "operator": "+"}, "allRequiredParamsPresent": True, "fulfillmentText": "Sorry, there seems to be service error. Your request can't be completed", "fulfillmentMessages": [{"text": {"text": ["Sorry, there seems to be service error. Your request can't be completed"]}}], "outputContexts": [{"name": "projects/fake-project-id/agent/sessions/fake-session-id/contexts/__system_counters__", "parameters": {"no-input": 0.0, "no-match": 0.0, "first_operand": 2.0, "first_operand.original": "two", "second_operand": 2.0, "second_operand.original": "2", "operator": "+", "operator.original": "plus"}}], "intent": {"name": "projects/fake-project-id/agent/intents/8e7dd332-0500-11ec-8d58-71988d2ea5e7", "displayName": "calculator.SolveMathOperation"}, "intentDetectionConfidence": 1.0, "languageCode": "en"}, "originalDetectIntentRequest": {"source": "DIALOGFLOW_CONSOLE", "payload": {}}, "session": "projects/fake-project-id/agent/sessions/fake-session-id"}

def _get_mock_example_agent():
    class MockExampleAgent(Agent):
        """
        Real example agent may change
        """
        languages = ["en"]

    return MockExampleAgent

@patch("intents.connectors.dialogflow_es.connector.resolve_credentials")
def test_calculator_is_fulfilled(self):
    class MockCalculatorOperator(Entity):
        name="CalculatorOperator"

        __entity_language_data__ = {
            LanguageCode.ENGLISH: [
                language.EntityEntry("*", ["multiplied by", "times"]),
                language.EntityEntry("-", ['minus']),
            ]
        }

    @dataclass
    class MockSolveMathOperation(Intent):
        _test_call_count = 0
        name = "calculator.SolveMathOperation"

        first_operand: Sys.Integer
        second_operand: Sys.Integer
        operator: MockCalculatorOperator

        def fulfill(self, context, *args, **kwargs):
            MockSolveMathOperation._test_call_count += 1
            assert self.first_operand == 2
            assert self.second_operand == 2
            assert self.operator == '+'
            return MockSolveMathOperationResponse(4)

    @dataclass
    class MockSolveMathOperationResponse(Intent):
        name = "calculator.SolveMathOperationResponse"

        operation_result: Sys.Integer

    MockExampleAgent = _get_mock_example_agent()

    with patch('intents.language.intent_language_data'):
        MockExampleAgent.register(MockSolveMathOperation)
        MockExampleAgent.register(MockSolveMathOperationResponse)

    df = DialogflowEsConnector('/fake/path/to/credentials.json', MockExampleAgent)
    request = FulfillmentRequest(body=CALCULATOR_FULFILLMENT)
    response = df.fulfill(request)
    expected = {
        "followupEventInput": {
            "name": "E_CALCULATOR_SOLVE_MATH_OPERATION_RESPONSE",
            "languageCode": "en",
            "parameters": {
                "operation_result": 4
            }
        }
    }
    assert response == expected

    assert MockSolveMathOperation._test_call_count == 1



import json
from datetime import datetime

from intents.connectors._experimental.alexa import fulfillment_schemas as fs

def test_parse_docs_request():
    """
    This is the request at
    https://developer.amazon.com/en-US/docs/alexa/custom-skills/request-and-response-json-reference.html#request-body-syntax
    
    Request field has been replaced with actual IntentRequest
    """

    body = json.loads("""{
        "version": "1.0",
        "session": {
            "new": true,
            "sessionId": "amzn1.echo-api.session.[unique-value-here]",
            "application": {
            "applicationId": "amzn1.ask.skill.[unique-value-here]"
        },
        "attributes": {
            "key": "string value"
        },
        "user": {
            "userId": "amzn1.ask.account.[unique-value-here]",
            "accessToken": "Atza|AAAAAAAA...",
            "permissions": {
                "consentToken": "ZZZZZZZ..."
            }
        }
        },
        "context": {
            "System": {
            "device": {
                "deviceId": "string",
                "supportedInterfaces": {
                "AudioPlayer": {}
                }
            },
            "application": {
                "applicationId": "amzn1.ask.skill.[unique-value-here]"
            },
            "user": {
                "userId": "amzn1.ask.account.[unique-value-here]",
                "accessToken": "Atza|AAAAAAAA...",
                "permissions": {
                "consentToken": "ZZZZZZZ..."
                }
            },
            "person": {
                "personId": "amzn1.ask.person.[unique-value-here]",
                "accessToken": "Atza|BBBBBBB..."
            },
            "unit": {
                "unitId": "amzn1.ask.unit.[unique-value-here]",
                "persistentUnitId" : "amzn1.alexa.unit.did.[unique-value-here]"
            },
            "apiEndpoint": "https://api.amazonalexa.com",
            "apiAccessToken": "AxThk..."
            },
            "AudioPlayer": {
                "playerActivity": "PLAYING",
                "token": "audioplayer-token",
                "offsetInMilliseconds": 0
            }
        },
        "request": {
            "type": "IntentRequest",
            "requestId": "amzn1.echo-api.request.fake-request-id",
            "locale": "en-US",
            "timestamp": "2021-08-10T19:12:25Z",
            "intent": {
                "name": "smalltalk_hello",
                "confirmationStatus": "NONE"
            }
        }
    }""")

    expected = fs.FulfillmentBody(
        version="1.0",
        session=fs.FulfillmentSession(
            new=True,
            sessionId="amzn1.echo-api.session.[unique-value-here]",
            application=fs.FulfillmentSessionApplication(applicationId="amzn1.ask.skill.[unique-value-here]"),
            attributes={"key": "string value"},
            user=fs.FulfillmentSessionUser(
                userId="amzn1.ask.account.[unique-value-here]",
                accessToken="Atza|AAAAAAAA...",
                permissions={"consentToken": "ZZZZZZZ..."}
            )
        ),
        context=fs.FulfillmentContext(
            System=fs.FulfillmentContextSystem(
                apiAccessToken="AxThk...",
                apiEndpoint="https://api.amazonalexa.com",
                application=fs.FulfillmentContextSystemApplication(applicationId="amzn1.ask.skill.[unique-value-here]"),
                device=fs.FulfillmentContextSystemDevice(
                    deviceId="string",
                    supportedInterfaces={
                        "AudioPlayer": {}
                    }
                ),
                unit=fs.FulfillmentContextSystemUnit(
                    unitId="amzn1.ask.unit.[unique-value-here]",
                    persistentUnitId="amzn1.alexa.unit.did.[unique-value-here]"
                ),
                person=fs.FulfillmentContextSystemPerson(
                    personId="amzn1.ask.person.[unique-value-here]",
                    accessToken="Atza|BBBBBBB..."
                ),
                user=fs.FulfillmentContextSystemUser(
                    userId="amzn1.ask.account.[unique-value-here]",
                    accessToken="Atza|AAAAAAAA...",
                    permissions={
                       "consentToken": "ZZZZZZZ..."
                    }
                )
            ),
            AudioPlayer=fs.FulfillmentContextAudioPlayer(
                token="audioplayer-token",
                playerActivity=fs.PlayerActivity.PLAYING,
                offsetInMilliseconds=0
            )
        ),
        request=fs.FulfillmentIntentRequest(
            type=fs.RequestType.INTENT,
            requestId="amzn1.echo-api.request.fake-request-id",
            timestamp=datetime(year=2021, month=8, day=10, hour=19, minute=12, second=25),
            locale="en-US",
            intent=fs.FulfillmentIntentRequestIntent(
                name="smalltalk_hello",
                confirmationStatus=fs.IntentConfirmationStatus.NONE
            )
        )
    )

    assert fs.from_dict(body) == expected



from typing import Type

from intents import Intent
from intents.helpers.misc import camel_to_snake_case

def context_name(intent_cls: Type[Intent]) -> str:
    return "c_" + camel_to_snake_case(intent_cls.name.replace(".", "_")) # TODO: refine

def event_name(intent_cls: Type[Intent]) -> str:
    """
    Generate the default event name that we associate with every intent.

    >>> event_name('test.intent_name')
    'E_TEST_INTENT_NAME'
    """
    return "E_" + camel_to_snake_case(intent_cls.name.replace(".", "_")).upper()



import logging
from dataclasses import dataclass, field
from typing import Dict, List, Any, Type
from collections import defaultdict

from intents import Intent, Agent, LanguageCode, FulfillmentContext, FulfillmentResult
from intents.connectors.interface import deserialize_intent_parameters, Prediction, ServiceEntityMappings
from intents.language import intent_language
from intents.connectors._experimental.snips import prediction_format as f

logger = logging.getLogger(__name__)

@dataclass
class SnipsPrediction(Prediction):
    parse_result: f.ParseResult = field(default=None, repr=False)

class SnipsPredictionComponent:
    """
    This is the component of SnipsConnector that is responsible for handling
    prediction and trigger calls.
    """

    agent_cls: Type[Agent]
    entity_mappings: ServiceEntityMappings

    def __init__(self, agent_cls: Type[Agent], entity_mappings: ServiceEntityMappings):
        self.agent_cls = agent_cls
        self.entity_mappings = entity_mappings

    def intent_from_parse_result(self, parse_result: f.ParseResult) -> Intent:
        """
        Turn SnipsNLU output into an Intent class
        """
        intent_name = parse_result.intent.intentName
        intent_cls = self.agent_cls._intents_by_name.get(intent_name)
        if not intent_cls:
            raise ValueError(f"Snips returned intent with name '{intent_name}', but this was not found in the Agent "
                            f"definition. Make sure that the model is updated by running SnipsConnector.upload() again,"
                            f"and if the problem persists please open an issue on the Intents repository.")
        snips_parameters = _slot_list_to_param_dict(intent_cls, parse_result.slots)
        parameter_dict = deserialize_intent_parameters(snips_parameters, intent_cls, self.entity_mappings)
        return intent_cls(**parameter_dict)

    def prediction_from_parse_result(self, parse_result: f.ParseResult, lang: LanguageCode) -> SnipsPrediction:
        """
        Turn SnipsNLU output into a Prediction object
        """
        intent = self.intent_from_parse_result(parse_result)
        language_data = intent_language.intent_language_data(self.agent_cls, intent.__class__, lang)
        language_data = language_data[lang]
        fulfillment_messages, fulfillment_text = intent_language.render_responses(intent, language_data)
        return SnipsPrediction(
            intent=intent,
            confidence=parse_result.intent.probability,
            fulfillment_messages=fulfillment_messages,
            fulfillment_text=fulfillment_text,
            parse_result=parse_result
        )

    def prediction_from_intent(self, intent: Intent, lang: LanguageCode) -> SnipsPrediction:
        """
        This is used to generate prediction objects in trigger requests. Normally a
        trigger is a request to a prediction service, and the response is turned
        into a prediction. As Snips runs locally, we can just return the Intent
        object as it is received.

        Note that intent relations are not implemented yet. In the future, they must
        be solved as well.
        """
        language_data = intent_language.intent_language_data(self.agent_cls, intent.__class__, lang)
        language_data = language_data[lang]
        fulfillment_messages, fulfillment_text = intent_language.render_responses(intent, language_data)
        return SnipsPrediction(
            intent=intent,
            confidence=1.0,
            fulfillment_messages=fulfillment_messages,
            fulfillment_text=fulfillment_text,
            parse_result=None
        )

    def fulfill_local(self, prediction: SnipsPrediction, lang: LanguageCode, _stack: List[str]=None) -> SnipsPrediction:
        """
        Simulate the fulfillment flow in a local procedure. This method is called
        internally by :meth:`~SnipsConnector.predict` and
        :meth:`~SnipsConnector.trigger` to solve fulfillments before returning
        the Prediction to User.

        Fulfillments are solved **recursively**, but at the moment loops are
        not allowed: intent A can trigger intent B in fulfillment, but cannot
        trigger itself. Intent B can trigger every other intent but A and B, and
        so on.

        Args:
            prediction: A prediction object, as it is built before solving
                fulfillment
            lang: Language code of the prediction
            _stack: This is used internally to prevent loops in recursion
        Returns:
            The triggered intent that is returned by predicted
            :meth:`Intent.fulfill`. If :meth:`fulfill` is not defined, or if it
            returns `None`, return the input `prediction` object
        """
        if not _stack:
            _stack = []
    
        if not prediction.intent:
            logger.warning("Prediction contains no intent: %s", prediction)
            return prediction

        if prediction.intent.name in _stack:
            raise RecursionError("Circular fulfillment calls detected: intent '%s' is being "
                                 "fulfilled twice. Stack: %s. Make sure intents aren't fulfilled "
                                 "recursively. If this is intended, open a feature request issue "
                                 "on the Intents repository", prediction.intent.name, _stack)

        _stack.append(prediction.intent.name)

        context = FulfillmentContext(
            confidence=prediction.confidence,
            fulfillment_text=prediction.fulfillment_text,
            fulfillment_messages=prediction.fulfillment_messages,
            language=lang
        )
        fulfillment_result = FulfillmentResult.ensure(prediction.intent.fulfill(context))
        if fulfillment_result:
            if fulfillment_result.trigger:
                triggered = self.prediction_from_intent(fulfillment_result.trigger, lang)
                return self.fulfill_local(triggered, lang, _stack=_stack)
            else:
                logger.warning("Intent returned a fulfillment result without trigger. Trigger "
                               "is the only supported response in SnipsConnector. Other elements "
                               "will be ignored.")
        return prediction

def _slot_list_to_param_dict(
    intent_cls: Type[Intent],
    result_slots: List[f.ParseResultSlot]
) -> Dict[str, Any]:
    """
    Snips doesn't differentiate between list and non-list parameters. If an
    entity can be tagged multiple times, result slots will contain more than one
    match for that entity, even if the slot is not meant to be a list. So here we:

    * Collect matches per slot as lists
    * Pick first element for non-list parameters

    The resulting parameter dict can be fed to
    :func:`deserialize_intent_parameters` without the risk of raising exceptions
    """
    result_lists = defaultdict(list)
    for slot in result_slots:
        result_lists[slot.slotName].append(slot.value)

    result = {}
    schema = intent_cls.parameter_schema
    for slot_name, slot_values in result_lists.items():
        if slot_name not in schema:
            raise KeyError(f"Slot {slot_name} not found in intent {intent_cls} with schema: "
                        f"{schema}. Make sure your trained model is up to date with the "
                        "latest Agent definition. If it is, please file a bug in the Intents "
                        "repository")
        if schema[slot_name].is_list:
            result[slot_name] = slot_values
        else:
            if len(slot_values) > 1:
                logger.warning("Prediction returned more than one value for slot %s: %s. "
                            "Only the first value will be considered.")
            result[slot_name] = slot_values[0]
    return result



"""
This is a connector that can export an :class:`Agent` to the **Alexa** format.

.. warning::

    Alexa connector is **experimental**, it is published mainly to study the case and gather
    feedback: expect relevant rough edges. Also,

    * `predict` and `trigger` are not implemented (you don't call Alexa, Alexa calls you)
    * Entities :class:`Sys.Email` and :class:`Sys.Url` are not available in Alexa (regex entites aren't supported either, which makes it difficult to work around this one)
    * Intent relations are not considered

Official Request/Response schemas:
https://github.com/alexa/alexa-apis-for-python/tree/master/ask-sdk-model/ask_sdk_model
"""
import os
import json
import shutil
import logging

from intents import Agent, Intent
from intents.helpers.data_classes import to_dict
from intents.connectors.interface import Connector, ServiceEntityMappings, FulfillmentRequest
from intents.connectors._experimental.alexa import names, export, language, fulfillment, fulfillment_schemas
from intents.connectors._experimental.alexa.slot_types import ENTITY_MAPPINGS

logger = logging.getLogger(__name__)

class AlexaConnector(Connector):
    """
    This is an implementation of :class:`Connector` that enables Agents to
    work as Alexa projects.

    .. warning::

        This connector is **experimental**. Features may not be complete and
        behavior may change in next releases.

    At the moment, all you can do is to :meth:`export` an Agent in the Alexa
    format, and start the fulfillment development server with
    :func:`~intents.fulfillment.run_dev_server`. You will have to configure your
    Alexa agent from the console to hit your URL instead of its default lambda
    for fulfillment.
    """
    entity_mappings: ServiceEntityMappings = ENTITY_MAPPINGS

    invocation_name: str
    names_component: names.AlexaNamesComponent
    language_component: language.AlexaLanguageComponent
    fulfillment_component: fulfillment.AlexaFulfillmentComponent
    export_component: export.AlexaExportComponent

    def __init__(
        self,
        agent_cls: type(Agent),
        invocation_name: str,
        default_session: str=None,
        default_language: str="en"
    ):
        super().__init__(agent_cls, default_session=default_session,
                         default_language=default_language)
        self.invocation_name = invocation_name # TODO: model constraints
        self.names_component = names.AlexaNamesComponent(agent_cls)
        self.language_component = language.AlexaLanguageComponent(agent_cls)
        self.fulfillment_component = fulfillment.AlexaFulfillmentComponent(
            agent_cls,
            self.names_component,
            self.language_component
        )
        self.export_component = export.AlexaExportComponent(
            agent_cls,
            self.names_component,
            self.language_component,
            invocation_name
        )
    
    def export(self, destination: str):
        """
        Export Agent in the given folder:

        .. code-block:: python

            from example_agent.agent import ExampleAgent
            from intents.connectors._experimental.alexa import AlexaConnector

            alexa = AlexaConnector(ExampleAgent, "any invocation")
            alexa.export("./TMP_ALEXA")

        The export will generate one JSON file per language, they can be imported
        from the Alexa console. Destination will be overwritten if already existing.
        """
        rendered = self.export_component.render()

        if os.path.isdir(destination):
            logger.warning("Removing existing export folder: %s", destination)
            shutil.rmtree(destination)
        os.makedirs(destination)

        for lang, data in rendered.items():
            with open(os.path.join(destination, f"agent.{lang.value}.json"), "w") as f:
                json.dump(data, f, indent=4)

    def upload(self):
        """
        *Not implemented*
        """
        raise NotImplementedError()

    def predict(self, message: str, session: str = None, language: str = None) -> Intent:
        """
        *Not implemented*
        """
        raise NotImplementedError()

    def trigger(self, intent: Intent, session: str=None, language: str=None) -> Intent:
        """
        *Not implemented*
        """
        raise NotImplementedError()

    def fulfill(self, fulfillment_request: FulfillmentRequest) -> dict:
        logger.warning("Authentication for Alexa fulfillment requests is NOT supported. Do not use AexaConnector in production.")
        body_dict = fulfillment_request.body
        request_body = fulfillment_schemas.from_dict(body_dict)
        response_body = self.fulfillment_component.handle_fulfillment(request_body)
        result = to_dict(response_body)
        return result



import json
from unittest.mock import patch
from typing import List
from dataclasses import dataclass

import pytest

from intents import Intent, Entity, Agent, LanguageCode, Sys
from intents.language import EntityEntry
from intents.connectors._experimental.alexa import fulfillment, fulfillment_schemas, names, language

BASE_FULFILLMENT_BODY = json.loads("""{
    "version": "1.0",
    "session": {
        "new": false,
        "sessionId": "amzn1.echo-api.session.fake-session-id",
        "application": {
            "applicationId": "amzn1.ask.skill.fake-application-id"
        },
        "user": {
            "userId": "amzn1.ask.account.fake-user-id"
        }
    },
    "context": {
        "Viewports": [
            {
                "type": "APL",
                "id": "main",
                "shape": "RECTANGLE",
                "dpi": 213,
                "presentationType": "STANDARD",
                "canRotate": false,
                "configuration": {
                    "current": {
                        "mode": "HUB",
                        "video": {
                            "codecs": [
                                "H_264_42",
                                "H_264_41"
                            ]
                        },
                        "size": {
                            "type": "DISCRETE",
                            "pixelWidth": 1280,
                            "pixelHeight": 800
                        }
                    }
                }
            }
        ],
        "Viewport": {
            "experiences": [
                {
                    "arcMinuteWidth": 346,
                    "arcMinuteHeight": 216,
                    "canRotate": false,
                    "canResize": false
                }
            ],
            "mode": "HUB",
            "shape": "RECTANGLE",
            "pixelWidth": 1280,
            "pixelHeight": 800,
            "dpi": 213,
            "currentPixelWidth": 1280,
            "currentPixelHeight": 800,
            "touch": [
                "SINGLE"
            ],
            "video": {
                "codecs": [
                    "H_264_42",
                    "H_264_41"
                ]
            }
        },
        "Extensions": {
            "available": {
                "aplext:backstack:10": {}
            }
        },
        "System": {
            "application": {
                "applicationId": "amzn1.ask.skill.fake-application-id"
            },
            "user": {
                "userId": "amzn1.ask.account.fake-user-id"
            },
            "device": {
                "deviceId": "amzn1.ask.device.fake-device-id",
                "supportedInterfaces": {
                    "Alexa.Presentation.APL": {
                        "runtime": {
                            "maxVersion": "1.7"
                        }
                    }
                }
            },
            "apiEndpoint": "https://api.amazonalexa.com",
            "apiAccessToken": "aaA0aAAaAaAAA0AaAAAaaAaaAaAAAaA0AaAaAaaaAAA0AaAaaA.fake-access-token"
        }
    },
    "request": {}
}""")

class MockExampleAgent(Agent):
    """
    Real example agent may change
    """
    languages = ["en", "it"]

@dataclass
class ToyIntent(Intent):
    name = "ToyIntent"

class MockPizzaType(Entity):
    name="PizzaType"
    __entity_language_data__ = {
        LanguageCode.ENGLISH: [
            EntityEntry("Margherita", []),
            EntityEntry("Diavola", ['pepperoni']),
        ],
        LanguageCode.ITALIAN: [
            EntityEntry("Margherita", []),
            EntityEntry("Diavola", ['salamino piccante']),
        ]
    }

@dataclass
class MockOrderPizza(Intent):
    name = "restaurant.OrderPizza"

    pizza_type: MockPizzaType
    amount: Sys.Integer = 1

class MockCalculatorOperator(Entity):
    name="CalculatorOperator"

    __entity_language_data__ = {
        LanguageCode.ENGLISH: [
            EntityEntry("*", ["multiplied by", "times"]),
            EntityEntry("-", ['minus']),
        ],
        LanguageCode.ITALIAN: [
            EntityEntry("*", ["per", "moltiplicato"]),
            EntityEntry("-", ['meno']),
        ]
    }

@dataclass
class MockSolveMathOperation(Intent):
    name = "calculator.SolveMathOperation"

    first_operand: Sys.Integer
    second_operand: Sys.Integer
    operator: MockCalculatorOperator

@dataclass
class MockGreetFriends(Intent):
    name = "smalltalk.GreetFriends"

    friend_names: List[Sys.Person]

with patch('intents.language.intent_language_data'):
    MockExampleAgent.register(MockOrderPizza)
    MockExampleAgent.register(MockSolveMathOperation)
    MockExampleAgent.register(MockGreetFriends)
    MockExampleAgent.register(ToyIntent)

def _build_fulfillment_body_dict(request_dict: dict) -> dict:
    result = BASE_FULFILLMENT_BODY.copy()
    result["request"] = request_dict
    return result

def _get_fulfillment_component():
    names_component = names.AlexaNamesComponent(MockExampleAgent)
    language_component = language.AlexaLanguageComponent(MockExampleAgent)
    return fulfillment.AlexaFulfillmentComponent(
        MockExampleAgent,
        names_component,
        language_component
    )

def test_build_intent_single_parameter():
    # OrderPizza(pizza_type="Margherita")
    ORDER_PIZZA_REQUEST = json.loads("""{
            "type": "IntentRequest",
            "requestId": "amzn1.echo-api.request.45f9da6f-f39b-41b4-b5fd-46d0cbf875cd",
            "locale": "en-US",
            "timestamp": "2021-08-12T20:02:52Z",
            "intent": {
                "name": "restaurant_OrderPizza",
                "confirmationStatus": "NONE",
                "slots": {
                    "amount": {
                        "name": "amount",
                        "confirmationStatus": "NONE"
                    },
                    "pizza_type": {
                        "name": "pizza_type",
                        "value": "margherita",
                        "resolutions": {
                            "resolutionsPerAuthority": [
                                {
                                    "authority": "amzn1.er-authority.echo-sdk.amzn1.ask.skill.fake-application-id.PizzaType",
                                    "status": {
                                        "code": "ER_SUCCESS_MATCH"
                                    },
                                    "values": [
                                        {
                                            "value": {
                                                "name": "Margherita",
                                                "id": "PizzaType-Margherita"
                                            }
                                        }
                                    ]
                                }
                            ]
                        },
                        "confirmationStatus": "NONE",
                        "source": "USER",
                        "slotValue": {
                            "type": "Simple",
                            "value": "margherita",
                            "resolutions": {
                                "resolutionsPerAuthority": [
                                    {
                                        "authority": "amzn1.er-authority.echo-sdk.amzn1.ask.skill.fake-application-id.PizzaType",
                                        "status": {
                                            "code": "ER_SUCCESS_MATCH"
                                        },
                                        "values": [
                                            {
                                                "value": {
                                                    "name": "Margherita",
                                                    "id": "PizzaType-Margherita"
                                                }
                                            }
                                        ]
                                    }
                                ]
                            }
                        }
                    }
                }
            }
        }
    """)
    fulfillment_component = _get_fulfillment_component()

    fulfillment_body_dict = _build_fulfillment_body_dict(ORDER_PIZZA_REQUEST)
    body = fulfillment_schemas.from_dict(fulfillment_body_dict)
    intent = fulfillment_component.intent_from_fulfillment(body, lang=LanguageCode.ENGLISH)

    assert intent == MockOrderPizza(pizza_type="Margherita")

def test_build_intent_normalizes_entity_values():
    # SolveMathOperation(first_operand=4, second_operand=5, operator="*")
    #                                                                 ^ must be normalized
    CALCULATOR_REQUEST = json.loads("""{
            "type": "IntentRequest",
            "requestId": "amzn1.echo-api.request.077b28ef-9c2b-4373-a250-d70d5b90a0bd",
            "locale": "en-US",
            "timestamp": "2021-08-12T20:07:38Z",
            "intent": {
                "name": "calculator_SolveMathOperation",
                "confirmationStatus": "NONE",
                "slots": {
                    "first_operand": {
                        "name": "first_operand",
                        "value": "4",
                        "confirmationStatus": "NONE",
                        "source": "USER",
                        "slotValue": {
                            "type": "Simple",
                            "value": "4"
                        }
                    },
                    "second_operand": {
                        "name": "second_operand",
                        "value": "5",
                        "confirmationStatus": "NONE",
                        "source": "USER",
                        "slotValue": {
                            "type": "Simple",
                            "value": "5"
                        }
                    },
                    "operator": {
                        "name": "operator",
                        "value": "times",
                        "resolutions": {
                            "resolutionsPerAuthority": [
                                {
                                    "authority": "amzn1.er-authority.echo-sdk.amzn1.ask.skill.fake-application-id.CalculatorOperator",
                                    "status": {
                                        "code": "ER_SUCCESS_MATCH"
                                    },
                                    "values": [
                                        {
                                            "value": {
                                                "name": "multiplied by",
                                                "id": "CalculatorOperator-multipliedby"
                                            }
                                        }
                                    ]
                                }
                            ]
                        },
                        "confirmationStatus": "NONE",
                        "source": "USER",
                        "slotValue": {
                            "type": "Simple",
                            "value": "times",
                            "resolutions": {
                                "resolutionsPerAuthority": [
                                    {
                                        "authority": "amzn1.er-authority.echo-sdk.amzn1.ask.skill.fake-application-id.CalculatorOperator",
                                        "status": {
                                            "code": "ER_SUCCESS_MATCH"
                                        },
                                        "values": [
                                            {
                                                "value": {
                                                    "name": "multiplied by",
                                                    "id": "CalculatorOperator-multipliedby"
                                                }
                                            }
                                        ]
                                    }
                                ]
                            }
                        }
                    }
                }
            }
        }"""
    )
    fulfillment_component = _get_fulfillment_component()

    fulfillment_body_dict = _build_fulfillment_body_dict(CALCULATOR_REQUEST)
    body = fulfillment_schemas.from_dict(fulfillment_body_dict)
    intent = fulfillment_component.intent_from_fulfillment(body, lang=LanguageCode.ENGLISH)

    assert intent == MockSolveMathOperation(first_operand=4, second_operand=5, operator="*")

def test_list_slot_single_value():

    # Greet friends with a single friend
    LIST_SLOT_REQUEST__SINGLE_VALUE = json.loads("""{
        "type": "IntentRequest",
        "requestId": "amzn1.echo-api.request.fake-api-request",
        "locale": "en-US",
        "timestamp": "2021-08-11T21:46:11Z",
        "intent": {
            "name": "smalltalk_GreetFriends",
            "confirmationStatus": "NONE",
            "slots": {
                "friend_names": {
                    "name": "friend_names",
                    "value": "John",
                    "resolutions": {
                        "resolutionsPerAuthority": [
                            {
                                "authority": "AlexaEntities",
                                "status": {
                                    "code": "ER_SUCCESS_MATCH"
                                },
                                "values": [
                                    {
                                        "value": {
                                            "name": "John",
                                            "id": "https://ld.amazonalexa.com/entities/v1/1NaUgygHJx0BHxfQcKICqn"
                                        }
                                    },
                                    {
                                        "value": {
                                            "name": "John",
                                            "id": "https://ld.amazonalexa.com/entities/v1/5MPA8cqc69MBzaK3Z91KU9"
                                        }
                                    },
                                    {
                                        "value": {
                                            "name": "John",
                                            "id": "https://ld.amazonalexa.com/entities/v1/B2ebljbbk5pDAFEnHBPBXV"
                                        }
                                    },
                                    {
                                        "value": {
                                            "name": "John",
                                            "id": "https://ld.amazonalexa.com/entities/v1/HFaP3f1ENFoFfLQTbsC989"
                                        }
                                    },
                                    {
                                        "value": {
                                            "name": "John",
                                            "id": "https://ld.amazonalexa.com/entities/v1/uU6qXLKVIG9gK9jax0Lj00"
                                        }
                                    }
                                ]
                            }
                        ]
                    },
                    "confirmationStatus": "NONE",
                    "source": "USER",
                    "slotValue": {
                        "type": "Simple",
                        "value": "John",
                        "resolutions": {
                            "resolutionsPerAuthority": [
                                {
                                    "authority": "AlexaEntities",
                                    "status": {
                                        "code": "ER_SUCCESS_MATCH"
                                    },
                                    "values": [
                                        {
                                            "value": {
                                                "name": "John",
                                                "id": "https://ld.amazonalexa.com/entities/v1/1NaUgygHJx0BHxfQcKICqn"
                                            }
                                        },
                                        {
                                            "value": {
                                                "name": "John",
                                                "id": "https://ld.amazonalexa.com/entities/v1/5MPA8cqc69MBzaK3Z91KU9"
                                            }
                                        },
                                        {
                                            "value": {
                                                "name": "John",
                                                "id": "https://ld.amazonalexa.com/entities/v1/B2ebljbbk5pDAFEnHBPBXV"
                                            }
                                        },
                                        {
                                            "value": {
                                                "name": "John",
                                                "id": "https://ld.amazonalexa.com/entities/v1/HFaP3f1ENFoFfLQTbsC989"
                                            }
                                        },
                                        {
                                            "value": {
                                                "name": "John",
                                                "id": "https://ld.amazonalexa.com/entities/v1/uU6qXLKVIG9gK9jax0Lj00"
                                            }
                                        }
                                    ]
                                }
                            ]
                        }
                    }
                },
                "__Conjunction": {
                    "name": "__Conjunction",
                    "confirmationStatus": "NONE"
                }
            }
        }
    }""")
    fulfillment_component = _get_fulfillment_component()

    fulfillment_body_dict = _build_fulfillment_body_dict(LIST_SLOT_REQUEST__SINGLE_VALUE)
    body = fulfillment_schemas.from_dict(fulfillment_body_dict)
    with pytest.warns(None):
        intent = fulfillment_component.intent_from_fulfillment(body, lang=LanguageCode.ENGLISH)

    assert intent == MockGreetFriends(friend_names=["John"])

def test_list_slot_list_value():

    LIST_SLOT_REQUEST__LIST_VALUE = json.loads("""{
        "type": "IntentRequest",
        "requestId": "amzn1.echo-api.request.258ed402-09b7-46f1-82e0-b01102829aa3",
        "locale": "en-US",
        "timestamp": "2021-08-11T21:58:24Z",
        "intent": {
            "name": "smalltalk_GreetFriends",
            "confirmationStatus": "NONE",
            "slots": {
                "friend_names": {
                    "name": "friend_names",
                    "confirmationStatus": "NONE",
                    "source": "USER",
                    "slotValue": {
                        "type": "List",
                        "values": [
                            {
                                "type": "Simple",
                                "value": "Al John",
                                "resolutions": {
                                    "resolutionsPerAuthority": [
                                        {
                                            "authority": "AlexaEntities",
                                            "status": {
                                                "code": "ER_SUCCESS_MATCH"
                                            },
                                            "values": [
                                                {
                                                    "value": {
                                                        "name": "Al St. John",
                                                        "id": "https://ld.amazonalexa.com/entities/v1/1684nfEDrbFDPS5zG3J9Xh"
                                                    }
                                                },
                                                {
                                                    "value": {
                                                        "name": "Al. John",
                                                        "id": "https://ld.amazonalexa.com/entities/v1/15clLtXk5V8G1eOUXS7e6U"
                                                    }
                                                },
                                                {
                                                    "value": {
                                                        "name": "Con Covert",
                                                        "id": "https://ld.amazonalexa.com/entities/v1/FQThPE13ILbFNfD9kAkQSA"
                                                    }
                                                },
                                                {
                                                    "value": {
                                                        "name": "Kenny Clarke",
                                                        "id": "https://ld.amazonalexa.com/entities/v1/2VrKlUBYfd4CCfrQaYXwhe"
                                                    }
                                                },
                                                {
                                                    "value": {
                                                        "name": "Holy Rage^Judas Priest^Atkins/May Project^",
                                                        "id": "https://ld.amazonalexa.com/entities/v1/5weB5rEfNVqDuk8WQEVprY"
                                                    }
                                                }
                                            ]
                                        }
                                    ]
                                }
                            },
                            {
                                "type": "Simple",
                                "value": "jack",
                                "resolutions": {
                                    "resolutionsPerAuthority": [
                                        {
                                            "authority": "AlexaEntities",
                                            "status": {
                                                "code": "ER_SUCCESS_MATCH"
                                            },
                                            "values": [
                                                {
                                                    "value": {
                                                        "name": "Jack",
                                                        "id": "https://ld.amazonalexa.com/entities/v1/3RAHxMlM8kNCNynNc67lEJ"
                                                    }
                                                },
                                                {
                                                    "value": {
                                                        "name": "Jack",
                                                        "id": "https://ld.amazonalexa.com/entities/v1/59kZG7GNVruEiv0I7DMyyn"
                                                    }
                                                },
                                                {
                                                    "value": {
                                                        "name": "Jack",
                                                        "id": "https://ld.amazonalexa.com/entities/v1/63nC6NdpAiDDfhWh3rH97o"
                                                    }
                                                },
                                                {
                                                    "value": {
                                                        "name": "Jack",
                                                        "id": "https://ld.amazonalexa.com/entities/v1/Aihij8TmUB7GGmc0NB2WAP"
                                                    }
                                                },
                                                {
                                                    "value": {
                                                        "name": "Jack",
                                                        "id": "https://ld.amazonalexa.com/entities/v1/IXkcQEfhx2kE8yHeOy5O9T"
                                                    }
                                                }
                                            ]
                                        }
                                    ]
                                }
                            }
                        ]
                    }
                },
                "__Conjunction": {
                    "name": "__Conjunction",
                    "value": "and",
                    "confirmationStatus": "NONE",
                    "source": "USER",
                    "slotValue": {
                        "type": "Simple",
                        "value": "and"
                    }
                }
            }
        }
    }""")
    fulfillment_component = _get_fulfillment_component()

    fulfillment_body_dict = _build_fulfillment_body_dict(LIST_SLOT_REQUEST__LIST_VALUE)
    body = fulfillment_schemas.from_dict(fulfillment_body_dict)
    intent = fulfillment_component.intent_from_fulfillment(body, lang=LanguageCode.ENGLISH)

    assert intent == MockGreetFriends(friend_names=["Al St. John", "Jack"])
    #                                                ^ that's what Alexa matched ..


#
# TODO
#

SLOTS_UNFILLED_REQUEST = json.loads("""{
    "type": "IntentRequest",
    "requestId": "amzn1.echo-api.request.8703689c-d01c-4362-8100-81e14e95087e",
    "locale": "en-US",
    "timestamp": "2021-08-11T21:44:34Z",
    "intent": {
        "name": "smalltalk_GreetFriends",
        "confirmationStatus": "NONE",
        "slots": {
            "friend_names": {
                "name": "friend_names",
                "confirmationStatus": "NONE"
            },
            "__Conjunction": {
                "name": "__Conjunction",
                "confirmationStatus": "NONE"
            }
        }
    }
}"""
)

UNMATCHED_CUSTOM_ENTITY_REQUEST = {
    'version': '1.0',
    'session': {'new': False, 'sessionId': 'amzn1.echo-api.session.fake-session-id', 'application': {'applicationId': 'amzn1.ask.skill.fake-application-id'}, 'user': {'userId': 'amzn1.ask.account.fake-account'}},
    'context': {
        'Viewports': [{'type': 'APL', 'id': 'main', 'shape': 'RECTANGLE', 'dpi': 213, 'presentationType': 'STANDARD', 'canRotate': False, 'configuration': {'current': {'mode': 'HUB', 'video': {'codecs': ['H_264_42', 'H_264_41']}, 'size': {'type': 'DISCRETE', 'pixelWidth': 1280, 'pixelHeight': 800}}}}],
        'Viewport': {'experiences': [{'arcMinuteWidth': 346, 'arcMinuteHeight': 216, 'canRotate': False, 'canResize': False}], 'mode': 'HUB', 'shape': 'RECTANGLE', 'pixelWidth': 1280, 'pixelHeight': 800, 'dpi': 213, 'currentPixelWidth': 1280, 'currentPixelHeight': 800, 'touch': ['SINGLE'], 'video': {'codecs': ['H_264_42', 'H_264_41']}},
        'Extensions': {'available': {'aplext:backstack:10': {}}},
        'System': {'application': {'applicationId': 'amzn1.ask.skill.fake-application-id'}, 'user': {'userId': 'amzn1.ask.account.fake-account'}, 'device': {'deviceId': 'amzn1.ask.device.fake-device-id', 'supportedInterfaces': {'Alexa.Presentation.APL': {'runtime': {'maxVersion': '1.7'}}}}, 'apiEndpoint': 'https://api.amazonalexa.com', 'apiAccessToken': 'fake-access-token'}
    },
    'request': {
        'type': 'IntentRequest',
        'requestId': 'amzn1.echo-api.request.fake-request-id',
        'locale': 'en-US',
        'timestamp': '2021-08-13T19:20:45Z',
        'intent': {
            'name': 'restaurant_OrderPizza',
            'confirmationStatus': 'NONE',
            'slots': {
                'amount': {
                    'name': 'amount',
                    'value': '2',
                    'confirmationStatus': 'NONE',
                    'source': 'USER',
                    'slotValue': {'type': 'Simple', 'value': '2'}
                },
                'pizza_type': {
                    'name': 'pizza_type',
                    'value': 'margherita pizzas',
                    'resolutions': {
                        'resolutionsPerAuthority': [{'authority': 'amzn1.er-authority.echo-sdk.amzn1.ask.skill.fake-application-id.PizzaType', 'status': {'code': 'ER_SUCCESS_NO_MATCH'}}]},
                        'confirmationStatus': 'NONE',
                        'source': 'USER',
                        'slotValue': {'type': 'Simple', 'value': 'margherita pizzas', 'resolutions': {'resolutionsPerAuthority': [{'authority': 'amzn1.er-authority.echo-sdk.amzn1.ask.skill.fake-application-id.PizzaType', 'status': {'code': 'ER_SUCCESS_NO_MATCH'}}]}}
                }
            }
        }
    }
}



"""
Patches :class:`~intents.model.entity.Sys.MusicGenre`

Sources:

* https://www.musicgenreslist.com/
* https://blogs.transparent.com/chinese/music-genres-in-chinese/

"""
from typing import List, Union

from intents import Entity, LanguageCode
from intents.language.entity_language import make_language_data

# TODO: this is very incomplete

class I_IntentsMusicGenre(Entity):
    __intents_internal__ = True
    __entity_language_data__ = {
        LanguageCode.ENGLISH: make_language_data(["Classic", "Alternative", "Blues", "Opera", "Country", "Bluegrass", "Soul", "Rockabilly", ["Progressive", "Prog"], ["Progressive rock", "Prog rock"], ["Progressive metal", "Prog metal"], "Rock", ["Rock 'n' Roll", "Rock and Roll"], "Hard Rock", "Garage Rock", "Grunge", ["Heavy Metal", "Metal"], "Dance", "Dubstep", "Electroswing", "House", "Techno", "Trance", "Electronic", ["Funk", "Funky"], "Rap", ["Hip Hop", "Hip-Hop"], "Swing", "Lounge", "Drum & Bass", "Folk", "Pop", "Jazz", "Gospel", "Reggaeton", ["Brit Pop", "Britpop", "British Pop"], "Reggae", ["R&B", "Rhythm & Blues", "Rhythm and Blues"], "Ska", "Surf"]),
        LanguageCode.ITALIAN: make_language_data(["Elettronica", ["Musica classica", "Classica"], "Alternative", "Blues", "Opera", "Country", "Bluegrass", "Soul", "Rockabilly", ["Progressive", "Prog"], ["Rock progressivo", "Progressive rock", "Prog rock"], ["Progressive metal", "Prog metal"], "Rock", ["Rock 'n' Roll", "Rock and Roll"], "Hard Rock", "Garage Rock", "Grunge", ["Heavy Metal", "Metal"], "Dance", "Disco", "Dubstep", "Electroswing", "House", "Techno", "Trance", ["Funk", "Funky"], "Rap", ["Hip Hop", "Hip-Hop"], "Swing", "Lounge", "Drum & Bass", "Folk", "Pop", "Jazz", "Gospel", "Reggaeton", ["Brit Pop", "Britpop", "British Pop"], "Reggae", ["R&B", "Rhythm & Blues", "Rhythm and Blues"], "Ska", "Surf"]),
        LanguageCode.DUTCH: make_language_data([["Elektronische", "Elektronische muziek"], ["Klassieke", "Klassieke muziek"], "Alternative", "Blues", "Opera", "Country", "Bluegrass", "Soul", "Rockabilly", ["Progressive", "Prog"], ["Rock progressivo", "Progressive rock", "Prog rock"], ["Progressive metal", "Prog metal"], "Rock", ["Rock 'n' Roll", "Rock and Roll"], "Hard Rock", "Garage Rock", "Grunge", ["Heavy Metal", "Metal"], "Dance", "Disco", "Dubstep", "Electroswing", "House", "Techno", "Trance", ["Funk", "Funky"], "Rap", ["Hip Hop", "Hip-Hop"], "Swing", "Lounge", "Drum & Bass", "Folk", "Pop", "Jazz", "Gospel", "Reggaeton", ["Brit Pop", "Britpop", "British Pop"], "Reggae", ["R&B", "Rhythm & Blues", "Rhythm and Blues"], "Ska", "Surf"]),
        LanguageCode.FRENCH: make_language_data([["Musique électronique", "Electronica"], ["Musique classique", "Classique"], "Alternative", "Blues", ["Opéra", "Opera"], "Country", "Bluegrass", "Soul", "Rockabilly", ["Musique progressive", "Progressive", "Prog"], ["Rock progressive", "Progressive rock", "Prog rock"], ["Metal Progressive", "Progressive metal", "Prog metal"], "Rock", ["Rock 'n' Roll", "Rock and Roll"], "Hard Rock", "Garage Rock", "Grunge", ["Heavy Metal", "Metal"], "Dance", "Disco", "Dubstep", "Electroswing", "House", "Techno", "Trance", ["Funk", "Funky"], "Rap", ["Hip Hop", "Hip-Hop"], "Swing", "Lounge", "Drum & Bass", "Folk", "Pop", "Jazz", "Gospel", "Reggaeton", ["Brit Pop", "Britpop", "British Pop"], "Reggae", ["R&B", "Rhythm & Blues", "Rhythm and Blues"], "Ska", "Surf"]),
        LanguageCode.GERMAN: make_language_data([["Elektronische Musik", "elektronische"], ["klassische Musik", "klassische"], ["Opernmusik", "Opern"], ["Progressive", "Prog"], ["Progressive rock", "Prog rock"], ["Progressive metal", "Prog metal"], "Alternative", "Blues", "Country", "Bluegrass", "Soul", "Rockabilly", "Rock", ["Rock 'n' Roll", "Rock and Roll"], "Hard Rock", "Garage Rock", "Grunge", ["Heavy Metal", "Metal"], "Dance", "Disco", "Dubstep", "Electroswing", "House", "Techno", "Trance", ["Funk", "Funky"], "Rap", ["Hip Hop", "Hip-Hop"], "Swing", "Lounge", "Drum & Bass", "Folk", "Pop", "Jazz", "Gospel", "Reggaeton", ["Brit Pop", "Britpop", "British Pop"], "Reggae", ["R&B", "Rhythm & Blues", "Rhythm and Blues"], "Ska", "Surf"]),
        LanguageCode.SPANISH: make_language_data([["Música electrónica", "Electrónica", "Electronic", "Electro"], ["Música clásica", "Clásica"], ["Opera", "Musica de opera"], ["Progressive", "Prog"], ["Rock progresivo", "Progressive rock", "Prog rock"], ["Metal progresivo", "Progressive metal", "Prog metal"], "Alternative", "Blues", "Country", "Bluegrass", "Soul", "Rockabilly", "Rock", ["Rock 'n' Roll", "Rock and Roll"], "Hard Rock", "Garage Rock", "Grunge", ["Heavy Metal", "Metal"], "Dance", "Disco", "Dubstep", "Electroswing", "House", "Techno", "Trance", ["Funk", "Funky"], "Rap", ["Hip Hop", "Hip-Hop"], "Swing", "Lounge", "Drum & Bass", "Folk", "Pop", "Jazz", "Gospel", "Reggaeton", ["Brit Pop", "Britpop", "British Pop"], "Reggae", ["R&B", "Rhythm & Blues", "Rhythm and Blues"], "Ska", "Surf"]),
        LanguageCode.CHINESE: make_language_data(["另类", "兰草", "蓝调", "古典", "乡村", "轻", "电子", "民乐", "重金属", "嘻哈", "爵士", "歌剧", "流行", "说唱", "节奏布鲁斯", "雷鬼", "摇滚", "灵魂", "传统", "世界"]),
    }



"""
Intent and Entity names must respect some constraints. :func:`check_name` documents
and implements those constraints.
"""
import re

def check_name(candidate_name: str, is_system: bool=False):
    """
    Raise `ValueError` if the given Intent or Entity name is not a valid name.
    Valid names

    * Only contain letter, underscore (`_`) and period (`.`) characters
    * Start with a letter
    * Don't contain repeated underscores (e.g. `__`)
    * Don't start wit `i_`. This is a reserved prefix for *Intents* system
      intents
    
    Note that `Agent.register` will apply further checks to spot duplicate
    names. Note that names are case insensitive, and shouldn't overlap
    with parameter names.

    Args:
        candidate_name: The Intent or Entity name to check
        is_system: When True, allow reserved names (those starting with `i_`)
    """
    invalid_reason = None
    
    if re.search(r'[^a-zA-Z_\.]', candidate_name):
        invalid_reason = "must only contain letters, underscore or period"

    if candidate_name.startswith('.') or candidate_name.startswith('_'):
        invalid_reason = "must start with a letter"

    if "__" in candidate_name:
        invalid_reason = "must not contain __"

    if candidate_name.lower().startswith("i_") and not is_system:
        invalid_reason = "the 'i_' prefix is reserved for system intents and entities"

    if invalid_reason:
        raise ValueError(f"Invalid name '{candidate_name}': {invalid_reason}. "
            "If the issue is related to your class name or path you can either change names to "
            "be compliant, or use a custom name by setting 'Intent.name' manually. See the "
            "documentation at https://intents.readthedocs.io/ for more information on intent "
            "naming rules.")



"""
This module is **deprecated**. Use :mod:`intents.connectors.interface` instead
"""
import warnings

warnings.warn("intents.service_connector is DEPRECATED and will be removed in v0.4.0. "
              "Use intents.connectors.interface instead", DeprecationWarning)

from intents.connectors.interface.connector import Connector
from intents.connectors.interface.entity_mappings import EntityMapping, StringEntityMapping, PatchedEntityMapping, ServiceEntityMappings, deserialize_intent_parameters
from intents.connectors.interface.fulfillment import FulfillmentRequest
from intents.connectors.interface.prediction import Prediction



from typing import Any

import pytest

from intents import Sys, LanguageCode, Entity
from intents.model.entity import SystemEntityMixin
from intents.resources.builtin_entities.color import I_IntentsColor
from intents.connectors.interface import entity_mappings

def test_string_entity_mapping():
    mapping = entity_mappings.StringEntityMapping(Sys.Integer, "@FakeInteger")
    assert mapping.entity_cls == Sys.Integer
    assert mapping.service_name == "@FakeInteger"
    assert mapping.from_service("42") == Sys.Integer("42")
    assert mapping.to_service(42) == "42"

def test_patched_entity_mapping():
    mapping = entity_mappings.PatchedEntityMapping(Sys.Color, I_IntentsColor)
    assert mapping.entity_cls == Sys.Color
    assert mapping.service_name == "I_IntentsColor"
    assert mapping.from_service("red") == Sys.Color("red")
    assert mapping.to_service(Sys.Color("red")) == "red"

#
# ServiceEntityMappings
#

class MockCustomMapping(entity_mappings.EntityMapping):
    entity_cls = Sys.Person
    service_name = "@FakePerson"
    supported_languages = [LanguageCode.ENGLISH]
    def from_service(self, service_data: Any) -> SystemEntityMixin:
        return Sys.Person(service_data["name"])
    def to_service(self, entity: SystemEntityMixin) -> Any:
        return str({"name": entity})

MOCK_MAPPINGS = entity_mappings.ServiceEntityMappings.from_list([
    entity_mappings.StringEntityMapping(Sys.PhoneNumber, "@FakePhoneNumber"),
    MockCustomMapping()
])

def test_mappings__sys_found():
    mapping = MOCK_MAPPINGS.lookup(Sys.PhoneNumber)
    assert mapping == entity_mappings.StringEntityMapping(Sys.PhoneNumber, "@FakePhoneNumber")
    
    assert MOCK_MAPPINGS.service_name(Sys.PhoneNumber) == "@FakePhoneNumber"
    assert MOCK_MAPPINGS.is_mapped(Sys.PhoneNumber, LanguageCode.ENGLISH)
    assert MOCK_MAPPINGS.is_mapped(Sys.PhoneNumber, LanguageCode.SPANISH_LATIN_AMERICA)

def test_mappings__supported_languages():
    mapping = MOCK_MAPPINGS.lookup(Sys.Person)
    assert isinstance(mapping, MockCustomMapping)
    
    assert MOCK_MAPPINGS.service_name(Sys.Person) == "@FakePerson"
    assert MOCK_MAPPINGS.is_mapped(Sys.Person, LanguageCode.ENGLISH)
    assert not MOCK_MAPPINGS.is_mapped(Sys.Person, LanguageCode.SPANISH_LATIN_AMERICA)

def test_mappings__sys_not_found():
    with pytest.raises(KeyError):
        MOCK_MAPPINGS.lookup(Sys.Color)

    assert not MOCK_MAPPINGS.is_mapped(Sys.Color, LanguageCode.ENGLISH)

def test_mappings__custom_entity():
    class MyCustomEntity(Entity):
        pass

    mapping = MOCK_MAPPINGS.lookup(MyCustomEntity)
    assert mapping == entity_mappings.StringEntityMapping(
        entity_cls=MyCustomEntity,
        service_name="MyCustomEntity"
    )
    assert MOCK_MAPPINGS.service_name(MyCustomEntity) == "MyCustomEntity"



from enum import Enum
from dataclasses import dataclass, asdict

from intents.helpers.data_classes import custom_asdict_factory, OmitNone

def test_custom_asdict_enums_are_converted():
    class ToyEnum(Enum):
        ONE = "one"
        TWO = "two"

    @dataclass
    class ToyDataclass:
        foo: str
        enum: ToyEnum

    dc = ToyDataclass("bar", ToyEnum.TWO)
    expected = {
        "foo": "bar",
        "enum": "two"
    }

    assert asdict(dc, dict_factory=custom_asdict_factory()) == expected

def test_custom_asdict_omit_none():
    @dataclass
    class ToyDataclass:
        foo: str = OmitNone()

    dc = ToyDataclass("bar")
    expected = {
        "foo": "bar",
    }
    assert asdict(dc, dict_factory=custom_asdict_factory()) == expected

    dc = ToyDataclass(None)
    expected = {
        "foo": None,
    }
    assert asdict(dc, dict_factory=custom_asdict_factory()) == expected

    dc = ToyDataclass()
    expected = {}
    assert asdict(dc, dict_factory=custom_asdict_factory()) == expected

def test_custom_asdict_omit_none_multiple():
    @dataclass
    class ToyDataclassOne:
        foo: str = OmitNone()

    @dataclass
    class ToyDataclassTwo:
        bar: str = OmitNone()

    dc = ToyDataclassOne()
    expected = {}
    assert asdict(dc, dict_factory=custom_asdict_factory()) == expected

    dc = ToyDataclassTwo()
    expected = {}
    assert asdict(dc, dict_factory=custom_asdict_factory()) == expected



from intents.connectors._experimental.snips import prediction_format as pf

def test_from_dict_no_slots():
    parse_result = {
        'input': 'I want a coffee',
        'intent': {'intentName': 'AskCoffee', 'probability': 0.65},
        'slots': []
    }
    expected = pf.ParseResult(
        input="I want a coffee",
        intent=pf.ParseResultIntent(
            intentName="AskCoffee",
            probability=0.65
        ),
        slots=[]
    )
    assert pf.from_dict(parse_result) == expected

def test_from_dict_slots():
    parse_result = {
        'input': 'I want a dark roast espresso',
        'intent': {'intentName': 'AskEspresso', 'probability': 0.56},
        'slots': [
            {
                'range': {'start': 9, 'end': 13},
                'rawValue': 'dark',
                'value': {'kind': 'Custom', 'value': 'dark'},
                'entity': 'CoffeeRoast',
                'slotName': 'roast'
            }
        ]
    }
    expected = pf.ParseResult(
        input="I want a dark roast espresso",
        intent=pf.ParseResultIntent(
            intentName="AskEspresso",
            probability=0.56
        ),
        slots=[
            pf.ParseResultSlot(
                range=pf.ParseResultSlotRange(start=9, end=13),
                rawValue="dark",
                value={'kind': 'Custom', 'value': 'dark'},
                entity="CoffeeRoast",
                slotName="roast"
            )
        ]
    )
    assert pf.from_dict(parse_result) == expected



"""
Here we implement fulfillment handling. This module is responsible for parsing
requests from Alexa, build Intent objects, solve relations and run fulfillment
procedures.
"""
import logging
from dataclasses import dataclass, field
from typing import Dict, List, Any, Type

from intents import Intent, Agent, LanguageCode, FulfillmentContext, FulfillmentResult
from intents.connectors.interface import deserialize_intent_parameters, Prediction
from intents.language import intent_language
from intents.connectors._experimental.alexa import fulfillment_schemas as fs
from intents.connectors._experimental.alexa import names, slot_types, language

logger = logging.getLogger(__name__)

@dataclass
class AlexaPrediction(Prediction):
    fulfillment_request: fs.FulfillmentBody = field(default=None, repr=False)

class AlexaFulfillmentComponent:
    """
    This is the component of AlexaConnector that is responsible for handling
    fulfillment requests and... TODO: complete.
    """

    agent_cls: Type[Agent]
    names_component: names.AlexaNamesComponent
    language_component: language.AlexaLanguageComponent

    def __init__(
        self,
        agent_cls: Type[Agent],
        names_component: names.AlexaNamesComponent,
        language_component: language.AlexaLanguageComponent
    ):
        self.agent_cls = agent_cls
        self.names_component = names_component
        self.language_component = language_component

    def handle_fulfillment(self, request_body: fs.FulfillmentBody) -> fs.FulfillmentResponseBody:
        if request_body.request.type == fs.RequestType.LAUNCH:
            return _make_speech_response("Skill launched!")

        if request_body.request.type == fs.RequestType.SESSION_ENDED:
            return fs.FulfillmentResponseBody()

        locale = request_body.request.locale
        lang = self.language_component.alexa_locale_to_agent_language(locale)
        intent = self.intent_from_fulfillment(request_body, lang)
        result_text = self.fulfill_local(intent, lang)
        return _make_speech_response(result_text)

    def fulfill_local(
        self,
        intent: Intent,
        lang: LanguageCode,
        _stack: List[str]=None
    ) -> str:
        """
        Resolve recursive fulfillment triggers locally, 
        """
        if not _stack:
            _stack = []

        if intent.name in _stack:
            raise RecursionError("Circular fulfillment calls detected: intent '%s' is being "
                                 "fulfilled twice. Stack: %s. Make sure intents aren't fulfilled "
                                 "recursively. If this is intended, open a feature request issue "
                                 "on the Intents repository", intent.name, _stack)

        _stack.append(intent.name)

        language_data = intent_language.intent_language_data(self.agent_cls, intent.__class__, lang)[lang]
        rendered_messages, rendered_plaintext = intent_language.render_responses(intent, language_data)
        context = FulfillmentContext(
            confidence=1.0,
            fulfillment_text=rendered_plaintext,
            fulfillment_messages=rendered_messages,
            language=lang
        )
        fulfillment_result = FulfillmentResult.ensure(intent.fulfill(context))

        if fulfillment_result:
            if fulfillment_result.trigger:
                return self.fulfill_local(fulfillment_result.trigger, lang, _stack=_stack)
            else:
                logger.warning("Intent returned a fulfillment result without trigger. Trigger "
                               "is the only supported response in SnipsConnector. Other elements "
                               "will be ignored.")

        return rendered_plaintext

    def intent_from_fulfillment(self, request_body: fs.FulfillmentBody, lang: LanguageCode) -> Intent:
        alexa_intent_name = request_body.request.intent.name
        intent_name = self.names_component.alexa_to_intent_name(alexa_intent_name)
        intent_cls = self.agent_cls._intents_by_name.get(intent_name)
        if not intent_cls:
            raise ValueError(f"Alexa returned intent with name '{alexa_intent_name}', but this was not found in the Agent "
                             f"definition. Make sure that the cloud agent is up to date with model,"
                             f"and if the problem persists please open an issue on the Intents repository.")
        
        request_slots = list(request_body.request.intent.slots.values())
        alexa_parameters = self._slots_to_param_dict(intent_cls, request_slots, lang)
        parameter_dict = deserialize_intent_parameters(alexa_parameters, intent_cls, slot_types.ENTITY_MAPPINGS)
        return intent_cls(**parameter_dict)

    def fulfillment_result_to_response(
        self,
        fulfillment_result: FulfillmentResult,
        context: FulfillmentContext    
    ) -> fs.FulfillmentResponseBody:

        # No fulfillment -> return intent responses
        if not fulfillment_result:
            print("no result: Returning messages")
            result = fs.FulfillmentResponseBody(
                response=fs.FulfillmentResponse(
                    outputSpeech=fs.FulfillmentResponseOutputSpeech(
                        type=fs.OutputSpeechType.PlainText,
                        text=context.fulfillment_text
                    )
                )
            )
            print(result)
            return result

        # TODO: doesn't work
        # Fulfillment result -> trigger
        fulfillment_result = FulfillmentResult.ensure(fulfillment_result)
        if fulfillment_result.trigger:
            intent: Intent = fulfillment_result.trigger
            return fs.FulfillmentResponseBody(
                # TODO: set attributes to track context
                response=fs.FulfillmentResponse(
                    directives=fs.FulfillmentResponseDialogDelegateDirective(
                        updatedIntent=fs.FulfillmentResponseDialogDelegateUpdatedIntent(
                            name=self.names_component.intent_to_alexa(intent),
                            confirmationStatus=fs.IntentConfirmationStatus.NONE,
                            slots=self._slots_from_intent(intent)
                        )
                    )
                )
            )
        logger.warning("No trigger in fulfillment result: ignoring")

    def prediction_from_fulfillment(self, request_body: fs.FulfillmentBody):
        raise NotImplementedError()

    def _slots_to_param_dict(
        self,
        intent_cls: Type[Intent],
        request_slots: List[fs.IntentSlot],
        lang: LanguageCode
    ) -> Dict[str, Any]:
        parameter_schema = intent_cls.parameter_schema

        # TODO: map custom slot values with language.alexa_entry_id_to_value

        result = {}
        for slot in request_slots:
            slot: fs.IntentSlot

            # Slot not matched in utterance
            if not slot.slotValue:
                continue

            # System slots, like "__Conjunction"
            if slot.name.startswith("__"):
                continue

            if slot.name not in parameter_schema:
                raise ValueError(f"Alexa returned slot name '{slot.name}', but this is not defined in "
                                 f"Intent '{intent_cls}' with parameter schema: {parameter_schema}. "
                                 "Make sure that your cloud agent is up to date with your code, and "
                                 "if the problem persist please file an issue on the Intents repository/")

            if slot.slotValue.type == fs.SlotType.SIMPLE:
                value = self._get_simple_slot_value(slot.slotValue, lang)
                
                if parameter_schema[slot.name].is_list:
                    logger.warning("Parameter '%s.%s' is defined as list, but Alexa returned a single "
                                   "value ('%s'). Will be converted to list", intent_cls.name, slot.name,
                                   value)
                    value = [value]
            else:
                value = []
                for v in slot.slotValue.values:
                    if v.type != fs.SlotType.SIMPLE:
                        logger.warning("Slot value for parameter '%s.%s' contains nested lists. This is "
                                       "not supported, nested lists will be skipped.")
                        continue
                    value.append(self._get_simple_slot_value(v, lang))

                if not parameter_schema[slot.name].is_list:
                    logger.warning("Alexa returned a list value for parameter '%s.%s' ('%s'), but "
                                   "parameter is not defined as list. Only the first element will be "
                                   "returned", intent_cls.name, slot.name, value)
                    value = value[0]

            result[slot.name] = value

        return result

    def _get_simple_slot_value(
        self,
        slot_value: fs.IntentSlotValue,
        lang: LanguageCode
    ) -> Any:
        """
            # Amazon: meh...
            value = slot.value
            
            # Amazon: perfect
            value = slot.slotValue.resolutions.resolutionsPerAuthority[0].values[0].value.name
        """
        # Some entities have resolutions, and canonical value is hidden there
        if slot_value.resolutions:
            authority_resolution = slot_value.resolutions.resolutionsPerAuthority[0]
            
            # System entities are returned as they are
            if authority_resolution.authority == "AlexaEntities":
                return authority_resolution.values[0].value.name
            
            # Custom entities are looked up by id
            entry_id = authority_resolution.values[0].value.id
            return self.language_component.alexa_entry_id_to_value(entry_id, lang)

        # Some entities don't, and canonical value is at top level
        return slot_value.value

    def _slots_from_intent(self, intent: Intent) -> Dict[str, fs.IntentSlot]:
        result = {}
        parameter_schema = intent.parameter_schema
        for param_name, param_value in intent.parameter_dict().items():
            entity_cls = parameter_schema[param_name].entity_cls
            mapping = slot_types.ENTITY_MAPPINGS.lookup(entity_cls)
            result[param_name] = fs.IntentSlot(
                confirmationStatus=fs.IntentConfirmationStatus.NONE,
                name=param_name,
                value=mapping.to_service(param_value)
            )
        return result

def _make_speech_response(text: str) -> fs.FulfillmentResponseBody:
    return fs.FulfillmentResponseBody(
        response=fs.FulfillmentResponse(
            outputSpeech=fs.FulfillmentResponseOutputSpeech(
                type=fs.OutputSpeechType.PlainText,
                text=text
            ),
            shouldEndSession=False
        )
    )


"""
Here we define schemas and utilities to convert Dialogflow Response elements to
*Intents* native classes.

References:

* Webhooks Request/Response
    * https://cloud.google.com/dialogflow/es/docs/fulfillment-webhook#webhook_request
    * https://cloud.google.com/dialogflow/es/docs/reference/rpc/google.cloud.dialogflow.v2#webhookrequest
    * https://cloud.google.com/dialogflow/es/docs/fulfillment-webhook#webhook_response

TODO: consider using same schemas (e.g. for response messages) in Agent export
"""
from enum import Enum
from typing import Dict, List, Any
from dataclasses import dataclass, field
from collections import defaultdict
from abc import ABC, abstractmethod

import dacite
from google.cloud.dialogflow_v2 import types as df_types
from google.protobuf.json_format import MessageToDict

#
# Response Format
#

@dataclass
class QueryResultMessageText:
    text: List[str]

@dataclass
class QueryResultMessageImage:
    imageUri: str
    accessibilityTest: str = ""

@dataclass
class QueryResultMessageQuickReplies:
    quickReplies: List[str]
    title: str = ""

@dataclass
class QueryResultMessageCardButtons:
    text: str = ""
    postback: str = ""

@dataclass
class QueryResultMessageCard:
    title: str = ""
    subtitle: str = ""
    imageUri: str = ""
    buttons: List[QueryResultMessageCardButtons] = field(default_factory=list)

@dataclass
class QueryResultMessagePayload:
    payload: Dict[str, Any]

class QueryResultMessagePlatform(Enum):
    """
    Ref:
    https://cloud.google.com/dialogflow/es/docs/reference/rpc/google.cloud.dialogflow.v2#google.cloud.dialogflow.v2.Intent.Message.Platform
    """
    PLATFORM_UNSPECIFIED = "PLATFORM_UNSPECIFIED"
    FACEBOOK = "FACEBOOK"
    SLACK = "SLACK"
    TELEGRAM = "TELEGRAM"
    KIK = "KIK"
    SKYPE = "SKYPE"
    LINE = "LINE"
    VIBER = "VIBER"
    ACTIONS_ON_GOOGLE = "ACTIONS_ON_GOOGLE"
    GOOGLE_HANGOUTS = "GOOGLE_HANGOUTS"

@dataclass
class QueryResultMessage:
    """
    See
    https://cloud.google.com/dialogflow/es/docs/reference/rpc/google.cloud.dialogflow.v2#message
    """
    platform: QueryResultMessagePlatform = QueryResultMessagePlatform.PLATFORM_UNSPECIFIED
    text: QueryResultMessageText = None
    image: QueryResultMessageImage = None
    quickReplies: QueryResultMessageQuickReplies = None
    card: QueryResultMessageCard = None
    payload: QueryResultMessagePayload = None

@dataclass
class QueryResultContext:
    name: str
    lifespanCount: int = 0
    parameters: Dict[str, Any] = field(default_factory=dict)

    @property
    def simple_name(self):
        """
        Dialogflow will store a full context path in QueryResultContext.name.
        The last chunk is just the context name.
        """
        return self.name.split('/')[-1]

@dataclass
class QueryResultIntent:
    name: str
    displayName: str

@dataclass
class QueryResult:
    queryText: str
    languageCode: str
    parameters: Dict[str, Any] = None
    intent: QueryResultIntent = None
    intentDetectionConfidence: float = None
    fulfillmentText: str = ""
    fulfillmentMessages: List[QueryResultMessage] = field(default_factory=list)
    allRequiredParamsPresent: bool = False
    outputContexts: List[QueryResultContext] = field(default_factory=list)
    action: str = ""
    diagnosticInfo: dict = field(default_factory=dict)
    webhookPayload: dict = field(default_factory=dict) # TODO: model
    webhookSource: str = ''
    cancelsSlotFilling: bool = False
    speechRecognitionConfidence: float = None

class DetectIntentResponseWebhookStatus:
    code: int
    message: str
    details: List[Any] = field(default_factory=list)

@dataclass
class DetectIntentResponse:
    responseId: str
    queryResult: QueryResult
    webhookStatus: DetectIntentResponseWebhookStatus = None
    outputAudio: Any = None
    outputAudioConfig: dict = None # TODO: model

@dataclass
class WebhookRequest:
    session: str
    responseId: str
    queryResult: QueryResult
    originalDetectIntentRequest: dict = None

def from_dict(data_class: type, data: dict):
    """
    Wraps :func:`dacite.from_dict` configuring enums for Dialogflow Predictions

    Args:
        data_class: The dataclass to use as schema
        data: Will be converted into a dataclass instance
    """
    return dacite.from_dict(
        data_class=data_class,
        data=data,
        config=dacite.Config(cast=[QueryResultMessagePlatform])
    )

def from_protobuf(data_class: type, data):
    """
    Cast a protobuf structure to an equivalent dataclass type. We like dataclasses.
    """
    if hasattr(data, "_pb"):
        data_dict = MessageToDict(data._pb)
    else:
        data_dict = MessageToDict(data)
    return from_dict(data_class, data_dict)



"""
Here we manage webhook results
"""
import logging

from intents import FulfillmentResult, FulfillmentContext
from intents.language_codes import ensure_language_code
from intents.helpers.data_classes import to_dict
from intents.connectors.dialogflow_es.names import event_name
from intents.connectors.dialogflow_es import webhook_format as wf

logger = logging.getLogger(__name__)

def fulfillment_result_to_response(fulfillment_result: FulfillmentResult, context: FulfillmentContext) -> dict:
    if fulfillment_result.trigger:
        intent = fulfillment_result.trigger
        followup_event = wf.WebhookResponseFollowupEvent(
            name=event_name(intent),
            languageCode=ensure_language_code(context.language).value,
            parameters=intent.parameter_dict()
        )
        return to_dict(wf.WebhookResponse(
            followupEventInput=followup_event
        ))
    else:
        logger.warning("No trigger in fulfillment result. Triggers are the only fulfillment "
                       "response available in DialogflowEsConnector at the moment. Will send "
                       "empty response.")
    return {}



import os
import sys
import logging
from typing import List, Type

from intents.language_codes import LanguageCode, LANGUAGE_CODES, FALLBACK_LANGUAGE

logger = logging.getLogger(__name__)

def agent_language_folder(agent_cls: Type["intents.model.agent.Agent"]) -> str:
    main_agent_package_name = agent_cls.__module__.split('.')[0]
    main_agent_package = sys.modules[main_agent_package_name]
    if '__path__' not in main_agent_package.__dict__:
        # TODO: try workdir or something...
        logger.warning("Agent %s doesn't seem to be defined within a package. Language data will not be loaded.", agent_cls)
        return [], []

    agent_folder = main_agent_package.__path__[0]
    language_folder = os.path.join(agent_folder, 'language')
    if not os.path.isdir(language_folder):
        raise ValueError(f"No language folder found for agent {agent_cls} (expected: {language_folder})")

    return language_folder

def agent_supported_languages(agent_cls: Type["intents.model.agent.Agent"]) -> List[LanguageCode]:
    if agent_cls.languages:
        return agent_cls.languages

    result = []

    language_folder = agent_language_folder(agent_cls)
    for f in os.scandir(language_folder):
        if f.is_dir() and not f.name.startswith('.')  and not f.name.startswith('_'):
            if f.name in LANGUAGE_CODES:
                result.append(LanguageCode(f.name))
            else:
                logger.warning("Unrecognized language code: '%s' (must be one of %s). Skipping language data.", f.name, LANGUAGE_CODES)

    return result

def match_agent_language(agent_cls: Type["intents.model.agent.Agent"], language: LanguageCode) -> LanguageCode:
    """
    Return a Language Code among the ones supported by Agent that matches
    `language`.

    If Agent supports `language` directly, `language` is returned as it is.

    Otherwise, look for a viable fallback language (e.g.
    :class:`LanguageCode.ENGLISH` is a viable fallback for
    :class:`LanguageCode.ENGLISH_US`).

    Raise `KeyError` if there is no viable language in Agent that matches the
    input one.

    Args:
        agent_cls: An Agent class
        language: The Language code to match in Agent
    Returns:
        A language code that matches `language` and that is supported by Agent
    Raises:
        KeyError: If Agent doesn't support `language` or one of its fallbacks
    """
    # TODO: update export procedures to use this
    agent_languages = agent_supported_languages(agent_cls)
    if language in agent_languages:
        return language

    for fallback in FALLBACK_LANGUAGE[language]:
        if fallback in agent_languages:
            return fallback

    raise KeyError(f"Agent {agent_cls} does not support language {language}")



"""
Here we define dataclass schemas for the Snips NLU Agent format. These are used
to produce JSON Dataset that can be read by Snips NLU.

References:

* https://github.com/snipsco/snips-nlu/blob/master/snips_nlu_samples/sample_dataset.json
* https://snips-nlu.readthedocs.io/en/latest/dataset.html
"""
from typing import List, Dict, Union
from dataclasses import dataclass, field

import dacite
from intents.helpers.data_classes import OmitNone

#
# Intent
#

@dataclass
class DatasetIntentUtteranceTextSegment:
    text: str

@dataclass
class DatasetIntentUtteranceEntitySegment:
    text: str
    entity: str
    slot_name: str

@dataclass
class DatasetIntentUtterance:
    data: List[Union[DatasetIntentUtteranceEntitySegment, DatasetIntentUtteranceTextSegment]]

@dataclass
class DatasetIntent:
    utterances: List[DatasetIntentUtterance]

#
# Entity
#

@dataclass
class DatasetEntityEntry:
    value: str
    synonyms: List[str] = field(default_factory=list)

@dataclass
class DatasetEntity:
    use_synonyms: bool
    automatically_extensible: bool
    matching_strictness: float
    data: List[DatasetEntityEntry] = OmitNone()

#
# Dataset
#

@dataclass
class Dataset:
    intents: Dict[str, DatasetIntent]
    entities: Dict[str, Union[DatasetEntity, dict]]
    language: str

def from_dict(data: dict):
    return dacite.from_dict(Dataset, data)

# import json
# # from intents.connectors._experimental.snips.agent_format import from_dict
# a = from_dict(json.load(open('/home/dario/lavoro/snips-sandbox/agent.json')))


"""
Entities are special chunks of text in the User utterances, that are recognized
and extratcted as parameters during predictions. For instance, in an utterance like
*"My name is John"* we most likely want to match *"John"* as a Person Entity.

Two types of Entities exist:

* **System entities** are built-in types that most chatbot frameworks will
  recognize (names, cities, numbers, emails, ...). These are modelled in
  :class:`Sys`.
* **Custom entities** are user-defined entity types, that typically consist in a
  list of possible values and their synonims. For instance, a `PizzaType` custom entity
  will have values like *"margherita"*, *"marinara"*, *"diavola"* and so on, and each
  of these may define some synonims; this way, in an utterance like *"I want a
  pepperoni pizza"*, the *"pepperoni"* chunk can be recognized as a `PizzaType`
  parameter, and mapped to its correct name, which is *"diavola"*. Custom entities
  are defined by extending the :class:`Entity` base class.
"""
import datetime
from typing import Any, List, Dict
from dataclasses import dataclass

from intents.model import names

class EntityType(type):

    name: str = None
    __entity_language_data__: Dict["language.LanguageCode", List["language.EntityEntry"]] = None

    def __new__(cls, name, bases, dct):
        result_cls = super().__new__(cls, name, bases, dct)

        # Do not process Intent base class
        if name in ['Entity', 'EntityMixin', 'SystemEntityMixin', 'BuiltinEntity']:
            # TODO: check if user isn't defining "class Entity(intents.Entity)" for
            # some reason
            return result_cls

        if not result_cls.name:
            result_cls.name = result_cls.__name__

        is_system = hasattr(result_cls, "__intents_internal__")
        names.check_name(result_cls.name, is_system)

        return result_cls

    @property
    def metadata(cls) -> "Entity.Meta":
        """
        TODO: consider removing this altogether, a `metadata` field in Entity
        may be enough
        """
        meta = cls.__dict__.get('meta')
        if not meta:
            return Entity.Meta()

        return meta

class EntityMixin(metaclass=EntityType):
    """
    This is a mixin class for entities, that adds :meth:`from_df_response`, to
    build the Entity object from the match data in Dialogflow Responses.
    """

    @classmethod
    def from_df_response(cls, match: Any):
        """
        Buid the Entity object from the match data from a Dialogflow Response.
        Specifically, `match` is the :class:`dict` version of
        `queryResult.parameters.<PARAMETER_NAME>`.
        """
        if isinstance(match, dict):
            return cls(**match)

        return cls(match)

class Entity(str, EntityMixin):
    """
    Custom Entities are defined by users to match parameters that are specific
    to the Agent's domain. This is done by extending this class:

    .. code-block:: python

        from dataclasses import dataclass
        from intents import Intent, Entity

        class PizzaType(Entity):
            \"\"\"One of the pizza types that a Customer can order\"\"\"

        @dataclass
        class CustomerOrdersPizza(Intent):
            \"\"\"A little docstring for my Intent\"\"\"

            pizza_type: PizzaType

    Language resources are expected for the `PizzaType` Entity. Like Intents,
    these will be looked up from the folder where the :class:`Agent` main class
    is defined, and specifically in
    `language/<LANGUAGE-CODE>/ENTITY_PizzaType.yaml`. More details in
    :mod:`intents.language`.
    """

    name: str = None

    @dataclass
    class Meta:
        """
        This class is used to set user-controlled Entity parameters.

        .. warning::

            The use of Meta is strictly bound to Dialogflow; support
            may be dropped in the next releases of Intents

        Args:
            regex_entity: Whether to treat entity values as regular expressions
            automated_expansion: Service will try to match synonyms even if they
                are not explicitly declared
            fuzzy_matching: Service will ignore word order in matching
        """
        regex_entity: bool = False
        automated_expansion: bool = False
        fuzzy_matching: bool = False

    meta: Meta = None

class SystemEntityMixin(EntityMixin):
    """
    This is used to signal that the entity is one of the Dialogflow default
    entities and doesn't need language resources.
    """

class Sys:
    """
    This is a container class, that defines all the system entities that are
    supported in the *Intents* framework. You can use system entities just like
    this:
    
    .. code-block:: python

        from dataclasses import dataclass
        from intents import Intent, Sys

        @dataclass
        class UserSaysName(Intent):
            \"\"\"A little docstring for my Intent\"\"\"

            user_name: Sys.Person

    System entities inherit from Python base types, so

    .. code-block:: python

        >>> isinstance(Sys.Integer(42), int)
        True
        >>> from datetime import date
        >>> isinstance(Sys.Date.from_py_date(date(2021, 7, 30)), date)
        True

    **NOTE**: This class can be expanded to contain more sophisticated entities,
    such as time intervals, amounts with measurement units and such. These may
    not be natively supported on every prediction service: some attention is
    needed to ensure portability.
    """

    # class Any(str, SystemEntityMixin):
    #     """
    #     Matches any non-empty input
    #
    #     NOTE: Sys.Any is tricky: not all prediction services support it, and
    #     the ones that do (including Dialoglfow) often misbehave when using it.
    #     """

    class Date(datetime.date, SystemEntityMixin):
        """
        Matches a date as a Python :class:`datetime.date` object
        """

        @staticmethod
        def from_py_date(py_date: datetime.date):
            """
            Clone the given :class:`datetime.date` object into a `Sys.Date`
            object. This is mostly for internal use.

            Args:
                py_date: A Python date object
            """
            return Sys.Date(py_date.year, py_date.month, py_date.day)

    class Time(datetime.time, SystemEntityMixin):
        """
        Matches a time reference as a Python :class:`datetime.time` object
        """

        @staticmethod
        def from_py_time(py_time: datetime.time):
            """
            Clone the given :class:`datetime.time` object into a `Sys.Time`
            object. This is mostly for internal use.

            Args:
                py_time: A Python time object
            """
            return Sys.Time(py_time.hour, py_time.minute, py_time.second, tzinfo=py_time.tzinfo)

    class Integer(int, SystemEntityMixin):
        """
        Matches integers only
        """

    class Person(str, SystemEntityMixin):
        """
        Matches common given names, last names or their combinations.

        Note that in Dialogflow this is returned as an Object (e.g. `{"name":
        "John"}`), while here we define `Person` as a String. The Dialogflow
        module defines proper entity mappings to handle the conversion.
        """

    class Email(str, SystemEntityMixin):
        """
        Any standard email address
        """

    class PhoneNumber(str, SystemEntityMixin):
        """
        Any standard phone number
        """

    class Color(str, SystemEntityMixin):
        """
        Words describing colors
        """

    class Language(str, SystemEntityMixin):
        """
        Language names
        """

    class Url(str, SystemEntityMixin):
        """
        Any standard URL
        """

    class MusicArtist(str, SystemEntityMixin):
        """
        Matches the name of a music artist
        """

    class MusicGenre(str, SystemEntityMixin):
        """
        Matches a music genre (rock, pop, reggae, ...)
        """

    # @dataclass
    # class UnitCurrency(SystemEntityMixin):
    #     """
    #     Number + currency name
    #     """
    #     __df_name__ = "sys.unit-currency"

    #     amount: float
    #     currency: str



import logging
from enum import Enum
from typing import Union, Type
from datetime import datetime, date

from intents import Sys, Entity, EntityMixin
from intents.language import LanguageCode
from intents.resources.builtin_entities import color, language, music_genre, first_name
from intents.connectors.interface import EntityMapping, StringEntityMapping, PatchedEntityMapping, ServiceEntityMappings

logger = logging.getLogger(__name__)

class BuiltinEntityTypes(Enum):
    MusicArtist = "snips/musicArtist"
    Number = "snips/number"
    Date = "snips/date"

class SnipsStringEntityMapping(StringEntityMapping):
    """
    Entities are always predicted as dicts. Simple ones just have a "value"
    field inside.
    """
    def from_service(self, service_data: dict):
        return self.entity_cls(service_data["value"])

class SnipsPatchedEntityMapping(PatchedEntityMapping):
    """
    Entities are always predicted as dicts. Simple ones just have a "value"
    field inside.
    """
    def from_service(self, service_data: dict):
        return self.entity_cls(service_data["value"])

class PlaceholderEntityMapping(SnipsStringEntityMapping):
    """
    Some entities are not supported in Snips, but can be replaced by empty
    placeholder custom entities. Snips NLU will try to infer values based on
    example utterance data only.

    This class is only meant to signal the export procedure that when this
    mapping is needed, an empty entity should be created.
    """

# TODO: implement custom processing unit for Snips to support regex-based
# entities (url, email, phone number, ...)

class DateMapping(EntityMapping):
    """
    This is a sample prediction with a Date entity:

    .. code-block:: json

    {
        "input": "My birthday is in two weeks",
        "intent": {
            "intentName": "UserSaysBirthday",
            "probability": 1.0
        },
        "slots": [
            {
                "range": {
                    "start": 15,
                    "end": 27
                },
                "rawValue": "in two weeks",
                "value": {
                    "kind": "InstantTime",
                    "value": "2021-08-21 00:00:00 +02:00",
                    "grain": "Day",
                    "precision": "Exact"
                },
                "entity": "snips/date",
                "slotName": "birthday_date"
            }
        ]
    }

    The mapping will receive the content of `slots[0].value`
    """
    entity_cls = Sys.Date
    service_name = BuiltinEntityTypes.Date.value
    supported_languages = [LanguageCode.ENGLISH]

    def from_service(self, service_data: dict):
        if (kind := service_data.get("kind")) != 'InstantTime':
            logger.warning("Expected returned date to be of 'InstantTime' kind. Snips returned '%s' "
                           "instead. This may cause unpredictable behavior.", kind)

        if (grain := service_data.get("grain")) != 'Day':
            logger.warning("Expected returned date to be of 'Day' grain. Snips returned '%s' "
                           "instead. This may cause unpredictable behavior.", grain)

        dt = datetime.fromisoformat(service_data["value"])
        return Sys.Date.from_py_date(dt.date())

    def to_service(self, entity: Union[date, datetime, str]):
                                      # ^ Sys.Date is subclass of date
        if isinstance(entity, datetime):
            entity = entity.date()
        return str(entity)

class SnipsEntityMappings(ServiceEntityMappings):

    def custom_entity_mapping(self, entity_cls: Type[EntityMixin]):
        return SnipsStringEntityMapping(
            entity_cls=entity_cls,
            service_name=entity_cls.name
        )

ENTITY_MAPPINGS = SnipsEntityMappings.from_list([
    SnipsPatchedEntityMapping(Sys.Person, first_name.I_IntentsFirstName),
    SnipsPatchedEntityMapping(Sys.Color, color.I_IntentsColor),
    DateMapping(),
    PlaceholderEntityMapping(Sys.Email, "I_PlaceholderEmail"),
    SnipsStringEntityMapping(Sys.Integer, BuiltinEntityTypes.Number.value),
    PlaceholderEntityMapping(Sys.PhoneNumber, "I_PlaceholderPhoneNumber"),
    PatchedEntityMapping(Sys.Language, language.I_IntentsLanguage),
    PlaceholderEntityMapping(Sys.Url, "I_PlaceholderUrl"),
    SnipsStringEntityMapping(Sys.MusicArtist, BuiltinEntityTypes.MusicArtist.value),
    SnipsPatchedEntityMapping(Sys.MusicGenre, music_genre.I_IntentsMusicGenre),
])

BUILTIN_ENTITIES = [m.entity_cls for m in ENTITY_MAPPINGS.values() if m.service_name in [x.value for x in BuiltinEntityTypes]]
PATCHED_MAPPINGS = [m for m in ENTITY_MAPPINGS.values() if isinstance(m, PatchedEntityMapping)]
PLACEHOLDER_ENTITIES = [m.entity_cls for m in ENTITY_MAPPINGS.values() if isinstance(m, PlaceholderEntityMapping)]



"""
The interpretation of an utterance may change depending on the context of the
conversation. Context-dependant dynamics in the conversation are modeled by the
**relations** intents have with each other. This module defines an API to
specify them.

.. note::

   Usage of relations is demonstrated in the :mod:`example_agent.shop` module of
   the Example Agent

Inheritance
===========
Since we define intents as Python classes, we can reuse of one of the founding
concepts of object oriented programming: **inheritance**.

.. code-block:: python

    @dataclass
    class AskCoffee(Intent):
        \"\"\"Hello, I'd like a light roast coffee\"\"\"
        roast_level: CoffeeRoastLevel = "medium"

    @dataclass
    class AskEspresso(AskCoffee):
        \"\"\"Hello, I'd like a dark roast espresso.\"\"\"

An intent should subclass another when its meaning **includes and specifies**
the meaning of the other intent. When an intent is a subclass of another:

    * It will inherit all **parameters**, types and default values of the parent intent
    * It will inherit all the other **relations** of the parent intent
    * It will inherit the parent's **fulfillment** method
    * It will be rendered as a separate, independent intent

      * It will **not** inherit language data such as example utterances and responses

Follow
======

The :func:`follow` relation is a context constraint: an intent that follows another
one can only be predicted after the followed one.

.. code-block:: python

    from intents import follow

    @dataclass
    class AddMilk(Intent):
        \"\"\"With milk please\"\"\"
        parent_ask_coffee: AskCoffee = follow()

An intent should follow another when it only makes sense in the parent's
context. In the example, it doesn't make sense to walk into a cafè and utter
"With milk please"; however, it does make sense to say so after asking for a
coffee.

Context parameters can be accessed with another OOP fundamental concept:
**composition**.

.. code-block:: python

    >>> prediction = connector.predict("I want a dark roast espresso")
    >>> prediction = connector.predict("With milk please")
    >>> prediction.intent.parent_ask_coffee.roast_level
    "dark"

This relation is implemented by looking at the **lifespan** of intents.
`AskCoffee` starts with a lifespan of 5 (this can be configured by setting a
`lifespan = N` property in the parent Intent class). This value is decremented
at each conversation turn; intents that follow `AskCoffee` can only be predicted
while its lifespan is > 0. :func:`follow` can also replice the current lifespan
of an Intent with a new one (e.g. set it to 0 to kill the context), see its
documentation for details.

It's worth noting that the *follow* relation is **inherited** by subclasses:

* If intent `AskEspresso` is a subclass of `AskCoffee`, and `AddMilk` follows
  `AskCoffe`, then `AddMilk` also follows `AskEspresso`.
* If intent `AskSkimmedMilk` is a subclass of `AskMilk`, and `AskMilk` follows
  `AskCoffee`, then `AskSkimmedMilk` also follows `AskCoffee` (and `AskEspresso`
  in the example above)

API
===
"""
from enum import Enum
from typing import List, Union, Type
import dataclasses
from dataclasses import dataclass, field

from intents import Intent

class RelationType(Enum):
    """
    Currently, the only available type is `RelationType.FOLLOW`. If you want to
    define a *follow* relation, use :func:`follow`.
    """
    FOLLOW = "follow"

@dataclass
class FollowRelationParameters:
    new_lifespan: int
    # fallback: Type[Intent]
    # within: int

def follow(*, new_lifespan: int=None) -> dataclasses.Field:
    """
    This can be used as a value for an Intent Relation field, e.g.

    .. code-block:: python
    
        from intents import follow
        
        @dataclass
        class AddMilk(Intent):
            \"\"\"With milk please\"\"\"
            parent_ask_coffee: AskCoffee = follow()

    This will make "AddMilk" predictable only if AskCoffe was predicted before,
    meaning, it is present in context with lifespan > 0. Lifespan is a counter
    that decreases from a given number (default is 5) at each conversation turn.
    :func:`follow` can update this number, either to keep the context alive, or
    to kill it when it's no more necessary:

    .. code-block:: python

        @dataclass
        class CancelAskCoffee(Intent):
            \"\"\"I don't want coffee anymore...\"\"\"
            parent_ask_coffee: AskCoffee = follow(new_lifespan=0)

    .. warning::

        The returned field currently sets `default=None` as a workaround to some
        known limitations of dataclasses with inheritance. This behavior may be
        adjusted again before 1.0

    Args:
        new_lifespan: Reset the lifespan of the followed intent to the given value
    """
    # TODO: solve inheritance after default fields and remove default
    return field(default=None, metadata={
        "relation_type": RelationType.FOLLOW,
        "relation_parameters": FollowRelationParameters(
            new_lifespan=new_lifespan
        )
    })

@dataclass
class IntentRelation:
    """
    Represent an Intent Relation definition from a "source" Intent (the one that
    defines the relation) to a "target" Intent (the one that is referenced in
    *source*).
    
    For instance, given

    .. code-block:: python

        @dataclass
        class AddMilk(Intent):
            \"\"\"With milk please\"\"\"
            parent_ask_coffee: AskCoffee = follow()

    Then the `follow` relation defined in `AddMilk` can be represented as

    .. code-block:: python

        IntentRelation(
            relation_type=RelationType.FOLLOW
            relation_parameters=None,
            field_name="parent_ask_coffee",
            source_cls=AddMilk,
            target_cls=AskCoffee
        )

    :class:`IntentRelation` objects are typically produced by
    :func:`intent_relations`, and enclosed in its returned
    :class:`IntentRelationMap` result structure.
    
    Args:
        relation_type: One of the relation types
        relation_parameters: Relation parameters as they are specified by the
            relation field
        field_name: Name of the field in `source_cls` that defines the relation
        source_cls: The Intent class where the relation is defined
        target_cls: The other Intent class that is referenced by the relation
    """
    relation_type: RelationType
    relation_parameters: type
    field_name: str
    source_cls: Type[Intent]
    target_cls: Type[Intent]

@dataclass
class FollowIntentRelation(IntentRelation):
    """
    Args:
        relation_parameters: Relation parameters as they are specified by the
            relation field
        field_name: Name of the field in `source_cls` that defines the relation
        source_cls: The Intent class where the relation is defined
        target_cls: The other Intent class that is referenced by the relation
    """
    relation_parameters: FollowRelationParameters
    relation_type: RelationType = field(default=RelationType.FOLLOW, init=None)

@dataclass
class IntentRelationMap:
    """
    A map of an Intent's relations, as it is produced by
    :func:`intent_relations`.

    Args:
        follow: A list of intents that are followed by the Relation subject
    """
    follow: List[FollowIntentRelation] = field(default_factory=list)

def intent_relations(intent: Union[Intent, Type[Intent]]) -> IntentRelationMap:
    """
    Produce a map of all the relations that are defined by the given Intent.

    Args:
        intent: The relation subject. Could be an Intent class or an Intent instance
    """
    intent_cls = intent if isinstance(intent, type) else type(intent)
    result = IntentRelationMap()
    for cls_field in intent.__dataclass_fields__.values():
        relation_type: RelationType = cls_field.metadata.get("relation_type")
        if relation_type:
            assert relation_type in [RelationType.FOLLOW]
            if relation_type == RelationType.FOLLOW:
                # TODO: check that there aren't other equivalent "follow" relations
                # (i.e. with same class or superclasses) in same intent
                relation_parameters: FollowRelationParameters = cls_field.metadata.get("relation_parameters")
                result.follow.append(FollowIntentRelation(
                    relation_parameters=relation_parameters,
                    field_name=cls_field.name,
                    source_cls=intent_cls,
                    target_cls=cls_field.type
                ))

    return result



from intents.language.entity_language import make_language_data, EntityEntry

def test_make_entity_language_string_entries():
    entries = ["red", "green", "blue"]
    result = make_language_data(entries)
    expected = [
        EntityEntry(value="red", synonyms=[]),
        EntityEntry(value="green", synonyms=[]),
        EntityEntry(value="blue", synonyms=[])
    ]
    assert result == expected

def test_make_entity_language_list_entries():
    entries = [
        ["red", "crimson", "scarlett"],
        ["green", "emerald"],
        ["blue", "cyan", "azure"]
    ]
    result = make_language_data(entries)
    expected = [
        EntityEntry(value="red", synonyms=["crimson", "scarlett"]),
        EntityEntry(value="green", synonyms=["emerald"]),
        EntityEntry(value="blue", synonyms=["cyan", "azure"])
    ]
    assert result == expected

def test_make_entity_language_mixed_entries():
    entries = [
        ["red", "crimson", "scarlett"],
        "green",
        ["blue", "cyan", "azure"]
    ]
    result = make_language_data(entries)
    expected = [
        EntityEntry(value="red", synonyms=["crimson", "scarlett"]),
        EntityEntry(value="green", synonyms=[]),
        EntityEntry(value="blue", synonyms=["cyan", "azure"])
    ]
    assert result == expected



"""
Connectors allow *Intents* Agent definitions to operate with real cloud services
such as Dialogflow, Lex or Azure Bot Services. Currently, only one stable connector is
provided with this library, and this is for Dialogflow ES:
:mod:`intents.connectors.dialogflow_es.connector`.

.. note::

    Details about the Connector interface are only useful if you intend to develop your own Service Connector (please consider raising a pull request if this is the case). If you just need to use the included Dialogflow Connector you can jump to its documentation page right away: :mod:`intents.connectors.dialogflow_es.connector`

Connectors are used to operate with the cloud version of the Agent, and
specifically to:

* Export an :class:`intents.Agent` in a format that is natively readable by the
  Service
* Predict User messages and trigger intents on the Cloud Agent
* Handle intent fulfillment requests by Service
"""
from intents.connectors.interface.connector import Connector
from intents.connectors.interface.entity_mappings import EntityMapping, \
    StringEntityMapping, PatchedEntityMapping, ServiceEntityMappings, deserialize_intent_parameters
from intents.connectors.interface.fulfillment import FulfillmentRequest, WebhookConfiguration
from intents.connectors.interface.prediction import Prediction
